import scipy.sparse as sp
import numpy as np
pos_per_user = {}
num_ratings=0
num_items=0
num_users=0
#user/item {raw id, inner id} map
userids = {}
itemids = {}
# inverse views of userIds, itemIds, 
idusers = {}
iditems={}
path="../Music"
train_matrix = None
social_matrix = None
time_matrix = None
with open(path+".rating", 'r') as f:
    for line in f.readlines():
        useridx, itemidx,rating, time= line.strip().split(",")
        
        num_ratings+=1
        if  itemidx not in itemids:
            iditems[num_items]=itemidx
            itemids[itemidx] = num_items
            num_items+=1

        if useridx not in userids:
            idusers[num_users]=useridx
            userids[useridx]=num_users
            num_users+=1
            pos_per_user[userids[useridx]]=[]
        pos_per_user[userids[useridx]].append((itemids[itemidx],rating,int(time)))
        # rating_matrix[self.userids[useridx],self.itemids[itemidx]] = rating
    print("\"num_users\": %d,\"num_items\":%d, \"num_ratings\":%d\n"%(num_users,num_items,num_ratings))
    train_matrix = sp.dok_matrix((num_users, num_items), dtype=np.int32)
    social_matrix = sp.dok_matrix((num_users, num_users), dtype=np.int32)
    time_matrix = sp.dok_matrix((num_users, num_items), dtype=np.int32)
    for u in np.arange(num_users):
        for enlement in pos_per_user[u]:
            train_matrix[u,enlement[0]]=enlement[1]
            time_matrix[u,enlement[0]] = enlement[2]
            
with open(path+".uu", 'r') as f:
    for line in f.readlines():
        useridx, socialidx = line.strip().split(",")
        if  useridx in userids and socialidx in userids:
            social_matrix[userids[useridx],userids[socialidx]]=1
            
f= open("rating.txt",'w')
f1= open("trust.txt",'w')

trainMatrix = train_matrix.tocsr()
socialMatrix = social_matrix.tocsr()
timeMatrix = time_matrix.tocsr()

for u in range(num_users):
    items_by_u = trainMatrix[u].indices
    social_by_u = socialMatrix[u].indices
    if len(items_by_u)>=10 and len(social_by_u)>=5:
        for i in items_by_u:
            rating = str(u)+","+str(i)+","+str(train_matrix[u,i])+","+str(time_matrix[u,i])+"\n"
            f.write(rating)
            
        for v in social_by_u:
            social = str(u)+","+str(v)+"\n"
            f1.write(social)

f1.close()            
f.close()           


'''
Reference: Dawen, Liang, et al. "Variational autoencoders for collaborative filtering." in WWW2018
@author: wubin
'''
import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner, tool
from tensorflow.contrib.layers import apply_regularization, l2_regularizer
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
from neurec.util.properties import Properties

class MultiVAE(AbstractRecommender):
    properties = [
        "learning_rate",
        "learner",
        "topk",
        "batch_size",
        "p_dim",
        "activation",
        "reg",
        "epochs",
        "verbose",
        "anneal_cap",
        "total_anneal_steps"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.batch_size= self.conf["batch_size"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items
        self.p_dims = self.conf["p_dim"] + [self.num_items]
        self.q_dims = self.p_dims[::-1]
        self.dims = self.q_dims + self.p_dims[1:]
        self.act = self.conf["activation"]
        self.reg = self.conf["reg"]
        self.num_epochs=self.conf["epochs"]
        self.loss_function="multinominal-likelihood"
        self.verbose=self.conf["verbose"]
        self.anneal_cap=self.conf["anneal_cap"]
        self.total_anneal_steps=self.conf["total_anneal_steps"]

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.input_ph = tf.placeholder(dtype=tf.float32, shape=[None, self.num_items])
            self.keep_prob_ph = tf.placeholder_with_default(1.0, shape=None)
            self.is_training_ph = tf.placeholder_with_default(0., shape=None)
            self.anneal_ph = tf.placeholder_with_default(1., shape=None)

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            self.weights_q, self.biases_q = [], []

            for i, (d_in, d_out) in enumerate(zip(self.q_dims[:-1], self.q_dims[1:])):
                if i == len(self.q_dims[:-1]) - 1:
                    # we need two sets of parameters for mean and variance,
                    # respectively
                    d_out *= 2
                weight_key = "weight_q_{}to{}".format(i, i+1)
                bias_key = "bias_q_{}".format(i+1)

                self.weights_q.append(tf.get_variable(
                    name=weight_key, shape=[d_in, d_out],
                    initializer=tf.contrib.layers.xavier_initializer()))

                self.biases_q.append(tf.get_variable(
                    name=bias_key, shape=[d_out],
                    initializer=tf.truncated_normal_initializer(
                        stddev=0.001)))


            self.weights_p, self.biases_p = [], []

            for i, (d_in, d_out) in enumerate(zip(self.p_dims[:-1], self.p_dims[1:])):
                weight_key = "weight_p_{}to{}".format(i, i+1)
                bias_key = "bias_p_{}".format(i+1)
                self.weights_p.append(tf.get_variable(
                    name=weight_key, shape=[d_in, d_out],
                    initializer=tf.contrib.layers.xavier_initializer()))

                self.biases_p.append(tf.get_variable(
                    name=bias_key, shape=[d_out],
                    initializer=tf.truncated_normal_initializer(
                        stddev=0.001)))
    def q_graph(self):
        mu_q, std_q, KL = None, None, None

        h = tf.nn.l2_normalize(self.input_ph, 1)
        h = tf.nn.dropout(h, self.keep_prob_ph)

        for i, (w, b) in enumerate(zip(self.weights_q, self.biases_q)):
            h = tf.matmul(h, w) + b

            if i != len(self.weights_q) - 1:
                h = tool.activation_function(self.act, h)
            else:
                mu_q = h[:, :self.q_dims[-1]]
                logvar_q = h[:, self.q_dims[-1]:]#log sigmod^2  batch x 200

                std_q = tf.exp(0.5 * logvar_q) #sigmod batch x 200
                KL = tf.reduce_mean(tf.reduce_sum(
                        0.5 * (-logvar_q + tf.exp(logvar_q) + tf.pow(mu_q,2) - 1), axis=1))
        return mu_q, std_q, KL

    def p_graph(self, z):
        self.h = z

        for i, (w, b) in enumerate(zip(self.weights_p, self.biases_p)):
            self.h = tf.matmul(self.h, w) + b

            if i != len(self.weights_p) - 1:
                self.h = tool.activation_function(self.act,self.h)
        return self.h

    def _create_inference(self):
        with tf.name_scope("inference"):
            # q-network
            mu_q, std_q, self.KL = self.q_graph()
            epsilon = tf.random_normal(tf.shape(std_q))

            sampled_z = mu_q + self.is_training_ph *\
                epsilon * std_q # batch x 200

            # p-network
            logits = self.p_graph(sampled_z)

            self.log_softmax_var = tf.nn.log_softmax(logits)


    def _create_loss(self):
        with tf.name_scope("loss"):
            neg_ll = -tf.reduce_mean(tf.reduce_sum(
            self.log_softmax_var * self.input_ph,
            axis=1))
            # apply regularization to weights
            reg = l2_regularizer(self.reg)

            reg_var = apply_regularization(reg, self.weights_q + self.weights_p)
            # tensorflow l2 regularization multiply 0.5 to the l2 norm
            # multiply 2 so that it is back in the same scale
            self.loss = neg_ll + self.anneal_ph * self.KL + 2 * reg_var #neg_ELBO

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()

    def train_model(self):
        update_count = 0.0
        # the total number of gradient updates for annealing
        # largest annealing parameter
        for epoch in  range(self.num_epochs):
            random_perm_doc_idx = np.random.permutation(self.num_users)
            self.total_batch = self.num_users
            total_loss = 0.0
            training_start_time = time()
            num_training_instances = self.num_users
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                if num_batch == self.total_batch - 1:
                    batch_set_idx = random_perm_doc_idx[num_batch * self.batch_size:]
                elif num_batch < self.total_batch - 1:
                    batch_set_idx = random_perm_doc_idx[num_batch * self.batch_size: (num_batch + 1) * self.batch_size]

                batch_matrix = np.zeros((len(batch_set_idx),self.num_items))

                if self.total_anneal_steps > 0:
                    anneal = min(self.anneal_cap, 1. * update_count / self.total_anneal_steps)
                else:
                    anneal = self.anneal_cap

                batch_uid = 0
                trainDict = self.dataset.trainDict
                for userid in batch_set_idx:
                    items_by_userid = trainDict[userid]
                    for itemid in items_by_userid:
                        batch_matrix[batch_uid,itemid] = 1

                    batch_uid=batch_uid+1

                feed_dict = feed_dict={self.input_ph: batch_matrix,self.keep_prob_ph: 0.5,
                            self.anneal_ph: anneal,self.is_training_ph: 1}
                _, loss = self.sess.run([self.optimizer, self.loss],feed_dict=feed_dict)
                total_loss+=loss

                update_count += 1
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        rating_matrix = np.zeros((1,self.num_items), dtype=np.int32)
        items_by_userid = self.dataset.trainDict[user_id]
        for itemid in items_by_userid:
            rating_matrix[0,itemid] = 1
        output = self.sess.run(self.h, feed_dict={self.input_ph:rating_matrix})
        return output[0,items]



from datetime import datetime
import logging
from neurec.data.properties import types
from neurec.data.models import models
from neurec.data.Dataset import Dataset
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties
from neurec.util import reader
from neurec.data.models import models
import numpy as np
import tensorflow as tf

properties = Properties()
dataset = {}

logger = logging.getLogger()
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(formatter)
logger.addHandler(ch)

ch = logging.FileHandler('neurec.log', mode='a')
ch.setLevel(logging.INFO)
ch.setFormatter(formatter)
logger.addHandler(ch)

def setup(properties_path, properties_section="DEFAULT", numpy_seed=2018, tensorflow_seed=2017):
    """Setups initial values for neurec.

    properties_path -- path to properties file
    properties_section -- section inside the properties files to read (default "DEFAULT")
    numpy_seed -- seed value for numpy random (default 2018)
    tensorflow_seed -- seed value for tensorflow random (default 2017)
    """
    np.random.seed(numpy_seed)
    tf.set_random_seed(tensorflow_seed)

    properties.setSection(properties_section)
    properties.setProperties(properties_path)

    data_input_path = properties.getProperty("data.input.path")
    dataset_name = properties.getProperty("data.input.dataset")
    splitter = properties.getProperty("data.splitter")
    separator = properties.getProperty("data.convert.separator")
    threshold = properties.getProperty("data.convert.binarize.threshold")
    evaluate_neg = properties.getProperty("rec.evaluate.neg")
    dataset_format = properties.getProperty("data.column.format")
    splitter_ratio = properties.getProperty("data.splitterratio")

    global dataset
    dataset = Dataset(data_input_path, dataset_name, dataset_format, splitter, separator, threshold, evaluate_neg, splitter_ratio)

def run():
    """Trains and evaluates a model."""
    logger = logging.getLogger(__name__)

    if not isinstance(dataset, Dataset):
       raise RuntimeError("Dataset not set. Call setup() function and pass a properties file to set the dataset")

    recommender = properties.getProperty("recommender")

    if not recommender in models:
        raise KeyError("Recommender " + str(recommender) + " not recognised. Add recommender to neurec.util.models")

    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    num_thread = properties.getProperty("rec.number.thread")

    with tf.Session(config=config) as sess:
        model = models[recommender](sess=sess)
        model.build_graph()
        sess.run(tf.global_variables_initializer())
        model.train_model()
        Evaluate.test_model(model, dataset, num_thread)

def listModels():
    """Returns a list of available models."""
    return models

def listProperties(model):
    """Returns a list of properties used by the model.

    name -- name of a model
    """
    model_properties = models[model].properties
    list = []

    for property in model_properties:
        list.append({
            "name": property,
            "type": types[property].__name__
        })

    return list



'''
Created on Aug 8, 2016
Processing datasets.

@author: Xiangnan He (xiangnanhe@gmail.com)
'''
import numpy as np
from neurec.data.LeaveOneOutDataSplitter import LeaveOneOutDataSplitter
from neurec.data.HoldOutDataSplitter import HoldOutDataSplitter
from neurec.data.GivenData import GivenData
from neurec.util.singleton import Singleton
from importlib import util
import os

class Dataset(metaclass=Singleton):

    def __init__(self, dataset_path, dataset_name, data_format, splitter,separator,threshold,evaluate_neg,splitterRatio=[0.8,0.2]):
        '''
        Constructor
        '''
        if (dataset_path == 'neurec'):
            neurec_path = util.find_spec('neurec', package='neurec').submodule_search_locations[0]
            dataset_path = os.path.join(neurec_path, 'dataset')
        self.path = dataset_path
        self.dataset_name = dataset_name
        self.data_format = data_format
        self.separator= separator
        self.threshold = threshold
        self.splitterRatio=splitterRatio
        self.evaluate_neg = evaluate_neg
        self.splitter=splitter
        self.num_users = 0
        self.num_items = 0
        self.trainMatrix = None
        self.trainDict =  None
        self.testMatrix =  None
        self.testNegatives =  None
        self.timeMatrix = None
        self.userseq = None
        self.userids = None
        self.itemids = None
        if splitter == "loo" :
            loo = LeaveOneOutDataSplitter(self.path, self.dataset_name, self.data_format,self.separator, self.threshold)
            self.trainMatrix,self.trainDict,self.testMatrix,\
            self.userseq,self.userids,self.itemids,self.timeMatrix = loo.load_data_by_user_time()
            self.num_users = self.trainMatrix.shape[0]
            self.num_items = self.trainMatrix.shape[1]
        elif splitter == "ratio" :
            hold_out = HoldOutDataSplitter(self.path, self.dataset_name,self.data_format, self.separator,self.threshold,self.splitterRatio)
            self.trainMatrix,self.trainDict,self.testMatrix,\
            self.userseq,self.userids,self.itemids,self.timeMatrix =\
            hold_out.load_data_by_user_time()
            self.num_users = self.trainMatrix.shape[0]
            self.num_items = self.trainMatrix.shape[1]
        elif splitter == "given":
            given = GivenData(self.path, self.dataset_name,self.separator,self.threshold)
            self.trainMatrix,self.trainDict,self.testMatrix,\
            self.userseq,self.userids,self.itemids,self.timeMatrix =\
            given.load_pre_splitter_data()
            self.num_users = self.trainMatrix.shape[0]
            self.num_items = self.trainMatrix.shape[1]
        else :
            print("please choose a splitter")

        self.testNegatives = self.get_negatives()
        self.num_users = self.trainMatrix.shape[0]
        self.num_items = self.trainMatrix.shape[1]

    def get_negatives(self):
        negatives = {}
        for u in np.arange(self.num_users):
            negative_per_user =[]
            if(self.evaluate_neg>0):
                for _ in np.arange(self.evaluate_neg):
                    neg_item_id = np.random.randint(0,self.num_items)
                    while (u,neg_item_id) in self.trainMatrix.keys() or  (u,neg_item_id) in self.testMatrix.keys() \
                          or neg_item_id in negative_per_user:
                        neg_item_id = np.random.randint(0, self.num_items)
                    negative_per_user.append(neg_item_id)
                negatives[u] = negative_per_user
                negative_per_user =[]
            else :
                negatives=None
        return  negatives



from neurec.evaluation.foldout import Precision 
from neurec.evaluation.foldout import Recall 
from neurec.evaluation.foldout import MAP 
from neurec.evaluation.foldout import MRR 
from neurec.evaluation.foldout import NDCG
import heapq  # for retrieval topK
import numpy as np
from concurrent.futures import ThreadPoolExecutor


def evaluate_by_foldout(model,evaluateMatrix,evaluateNegatives,num_thread):
    """
    Evaluate the performance (precision, recall, MAP,NDCG,MRR) of top-K recommendation
    Return: score of each test rating.
    """
    global _model
    global _trainMatrix
    global _testMatrix
    global _evaluateMatrix
    global _evaluateNegatives
    global _K
    global _evaluateusers
    _model = model
    _trainMatrix = _model.dataset.trainMatrix.tocsr()
    _testMatrix = _model.dataset.testMatrix.tocsr()
    _evaluateMatrix = evaluateMatrix.tocsr()
    _evaluateNegatives = evaluateNegatives
    _evaluateusers = []
    _K = _model.topK
    Pres, Recs,MAPs,NDCGs,MRRs = [],[],[],[],[]
    for u in range(_model.num_users):
        items_test = _testMatrix[u].indices
        if len(items_test) >0:
            _evaluateusers.append(u)
    if(num_thread > 1): # Multi-thread
        with ThreadPoolExecutor() as executor:
            res = executor.map(eval_by_foldout_user, _evaluateusers)
        #res = pool.map(eval_by_foldout_user, range(len(_evaluateMatrix)))
        #pool.close()
        #pool.join()
        res = list(res)
        Pres = [r[0] for r in res]
        Recs = [r[1] for r in res]
        MAPs = [r[2] for r in res]
        NDCGs = [r[3] for r in res]
        MRRs = [r[4] for r in res]
    # Single thread
    else:
        # Single thread
        for u in _evaluateusers:
            if len(_evaluateMatrix[u].indices) !=0:
                (Pre,Rec,MAP,NDCG,MRR) = eval_by_foldout_user(u)  
                Pres.append(Pre) 
                Recs.append(Rec)
                MAPs.append(MAP)
                NDCGs.append(NDCG)
                MRRs.append(MRR)
    return (Pres,Recs,MAPs,NDCGs,MRRs)
def eval_by_foldout_user(u):
    target_items= _evaluateMatrix[u].indices
    eval_items =[]
    if _evaluateNegatives != None:
        eval_items = _evaluateNegatives[u]
    else :
        all_items = set(np.arange(_model.num_items))
        eval_items = list(all_items - set(_trainMatrix[u].indices))
    eval_items.extend(target_items)
    # Get prediction scores
    map_item_score = {}
    predictions = _model.predict(u,eval_items)
    for i in np.arange(len(eval_items)):
        item = eval_items[i]
        map_item_score[item] = predictions[i]
    # Evaluate top rank list
    rank_list = heapq.nlargest(_K, map_item_score, key=map_item_score.get)
    target_items = set(target_items)
    Pre = Precision.getPre(rank_list, target_items,_K)
    Rec = Recall.getRec(rank_list, target_items)
    ap = MAP.getAP(rank_list, target_items)
    dcg = NDCG.getNDCG(rank_list, target_items)
    rr = MRR.getMRR(rank_list, target_items)
    return (Pre,Rec,ap,dcg,rr)


'''
Reference: Steffen Rendle et al., "Factorizing Personalized Markov Chains
for Next-Basket Recommendation." in WWW 2010.
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner, data_gen
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
class FPMCplus(AbstractRecommender):
    properties = [
        "learning_rate",
        "embedding_size",
        "weight_size",
        "learner",
        "loss_function",
        "ispairwise",
        "topk",
        "epochs",
        "reg_mf",
        "reg_w",
        "batch_size",
        "high_order",
        "verbose",
        "num_neg"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.embedding_size = self.conf["embedding_size"]
        self.weight_size = self.conf["weight_size"]
        self.learner = self.conf["learner"]
        self.loss_function = self.conf["loss_function"]
        self.ispairwise = self.conf["ispairwise"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.reg_mf = self.conf["reg_mf"]
        self.reg_w = self.conf["reg_w"]
        self.batch_size= self.conf["batch_size"]
        self.high_order = self.conf["high_order"]
        self.verbose= self.conf["verbose"]
        self.num_negatives= self.conf["num_neg"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape = [None,], name = "user_input")
            self.item_input = tf.placeholder(tf.int32, shape = [None,], name = "item_input_pos")
            self.item_input_recents = tf.placeholder(tf.int32, shape = [None,None], name = "item_input_recent")
            if self.ispairwise == True:
                self.item_input_neg = tf.placeholder(tf.int32, shape = [None,], name = "item_input_neg")
            else :
                self.lables = tf.placeholder(tf.float32, shape=[None,],name="labels")
    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.embeddings_UI = tf.Variable(tf.truncated_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01),
                name='embeddings_UI', dtype=tf.float32)  #(users, embedding_size)
            self.embeddings_IU = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='embeddings_IU', dtype=tf.float32)  #(items, embedding_size)
            self.embeddings_IL = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='embeddings_IL', dtype=tf.float32)
            self.embeddings_LI = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='embeddings_LI', dtype=tf.float32)  #(items, embedding_size)
            self.W = tf.Variable(tf.truncated_normal(shape=[3*self.embedding_size, self.weight_size],\
                mean=0.0, stddev=tf.sqrt(tf.div(2.0, self.weight_size + self.embedding_size))),\
                    name='Weights_for_MLP', dtype=tf.float32, trainable=True)
            self.b = tf.Variable(tf.truncated_normal(shape=[1, self.weight_size], mean=0.0, \
                stddev=tf.sqrt(tf.div(2.0, self.weight_size + self.embedding_size))),name='Bias_for_MLP', dtype=tf.float32, trainable=True)
            self.h = tf.Variable(tf.ones([self.weight_size, 1]), name='H_for_MLP', dtype=tf.float32)

    def _attention_MLP(self,embeddings_UI_u, embeddings_IL_i,item_embedding_recent):
        with tf.name_scope("attention_MLP"):
            UI_u = tf.tile(tf.expand_dims(embeddings_UI_u,1),tf.stack([1,self.high_order,1]))
            IL_i = tf.tile(tf.expand_dims(embeddings_IL_i,1),tf.stack([1,self.high_order,1]))
            embedding_short =tf.concat([UI_u,IL_i,item_embedding_recent],2)
            batch_users = tf.shape(embedding_short)[0]

            MLP_output = tf.matmul(tf.reshape(embedding_short,shape=[-1,3*self.embedding_size]), self.W) + self.b #(b*n, e or 2*e) * (e or 2*e, w) + (1, w)
            MLP_output = tf.nn.tanh(MLP_output)

            A_ = tf.reshape(tf.matmul(MLP_output, self.h),shape=[batch_users,self.high_order]) #(b*n, w) * (w, 1) => (None, 1) => (b, n)

            # softmax for not mask features
            exp_A_ = tf.exp(A_)
            exp_sum = tf.reduce_sum(exp_A_, 1, keepdims=True)  # (b, 1)

            A = tf.expand_dims(tf.div(exp_A_, exp_sum),2) # (b, n, 1)

            return tf.reduce_sum(A * item_embedding_recent, 1)

    def _create_inference(self, item_input):
        with tf.name_scope("inference"):
            # embedding look up
            embeddings_UI_u = tf.nn.embedding_lookup(self.embeddings_UI, self.user_input)
            embeddings_IU_i = tf.nn.embedding_lookup(self.embeddings_IU,item_input)
            embeddings_IL_i = tf.nn.embedding_lookup(self.embeddings_IL, item_input)
            embeddings_LI_l = tf.nn.embedding_lookup(self.embeddings_LI, self.item_input_recents)
            item_embedding_short = self._attention_MLP(embeddings_UI_u,embeddings_IL_i,embeddings_LI_l)
            predict_vector =  tf.multiply(embeddings_UI_u, embeddings_IU_i)+tf.multiply(embeddings_IL_i,item_embedding_short)
            predict = tf.reduce_sum(predict_vector, 1)
            return embeddings_UI_u,embeddings_IU_i, embeddings_IL_i,embeddings_LI_l,predict


    def _create_loss(self):
        with tf.name_scope("loss"):
            # loss for L(Theta)
            # loss for L(Theta)
            UI_u,IU_i,IL_i,LI_l,self.output = self._create_inference(self.item_input)
            if self.ispairwise == True:
                _, IU_j,IL_j,_,output_neg = self._create_inference(self.item_input_neg)
                self.result = self.output - output_neg
                self.loss = learner.pairwise_loss(self.loss_function,self.result) + self.reg_mf * ( tf.reduce_sum(tf.square(UI_u)) \
                + tf.reduce_sum(tf.square(IU_i)) + tf.reduce_sum(tf.square(IL_i)) + tf.reduce_sum(tf.square(LI_l))+ \
                 tf.reduce_sum(tf.square(LI_l))+tf.reduce_sum(tf.square(IU_j))+tf.reduce_sum(tf.square(IL_j)))+\
                 self.reg_w * (tf.reduce_sum(tf.square(self.W))+tf.reduce_sum(tf.square(self.h)))
            else :
                self.loss = learner.pointwise_loss(self.loss_function,self.lables,self.output) + self.reg_mf * (tf.reduce_sum(tf.square(UI_u)) \
                +tf.reduce_sum(tf.square(IU_i))+ tf.reduce_sum(tf.square(IL_i))+tf.reduce_sum(tf.square(LI_l)))

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()
#---------- training process -------
    def train_model(self):
        for epoch in  range(self.num_epochs):
            # Generate training instances
            if self.ispairwise == True:
                user_input, item_input_pos, item_input_recents, item_input_neg = \
                data_gen._get_pairwise_all_highorder_data(self.dataset,self.high_order)
            else :
                user_input, item_input,item_input_recents, lables = data_gen._get_pointwise_all_highorder_data(self.dataset,self.high_order, self.num_negatives)

            num_training_instances = len(user_input)
            total_loss = 0.0
            training_start_time = time()

            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                if self.ispairwise == True:
                    bat_users, bat_items_pos, bat_items_recents,bat_items_neg  = \
                    data_gen._get_pairwise_batch_seqdata(user_input, item_input_pos, \
                    item_input_recents, item_input_neg, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users,self.item_input:bat_items_pos,\
                                self.item_input_recents:bat_items_recents,self.item_input_neg:bat_items_neg}
                else :
                    bat_users, bat_items, bat_items_recents, bat_lables =\
                    data_gen._get_pointwise_batch_seqdata(user_input, \
                    item_input,item_input_recents, lables, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users, self.item_input:bat_items,
                                 self.item_input_recents:bat_items_recents,self.lables:bat_lables}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss

            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)
    def predict(self, user_id, items):
        cand_items = self.dataset.trainDict[user_id]
        item_recents = []
        for _ in range(len(items)):
            item_recents.append( cand_items[len(cand_items)-self.high_order:])
        users = np.full(len(items), user_id, dtype='int32')
        return self.sess.run((self.output), feed_dict={self.user_input: users,\
                                        self.item_input_recents:item_recents, self.item_input: items})



import math
def getNDCG(rank_list, target_items):
    dcg = 0
    idcg = IDCG(len(target_items))
    for i in range(len(rank_list)):
        item_id= rank_list[i]
        if (item_id not in target_items):
            continue
        rank = i +1
        dcg +=1/math.log(rank+1,2)
    return dcg/idcg

def IDCG(n):
    idcg = 0
    for i in range(n):
        idcg +=1/math.log(i+2,2)
    return idcg


def getAP(rank_list, target_items):
    hits = 0
    sum_precs = 0
    for n in range(len(rank_list)):
        if rank_list[n] in target_items:
            hits += 1
            sum_precs += hits / (n + 1.0)
    if hits > 0:
        Min = len(rank_list) if len(rank_list)<len(target_items) else len(target_items)
        return sum_precs / Min
    else:
        return 0


'''
Reference: Dawen, Liang, et al. "Variational autoencoders for collaborative filtering." in WWW2018
@author: wubin
'''
import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner, tool
from tensorflow.contrib.layers import apply_regularization, l2_regularizer
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
from neurec.util.properties import Properties

class MultiDAE(AbstractRecommender):
    properties = [
        "learning_rate",
        "learner",
        "topk",
        "batch_size",
        "p_dim",
        "activation",
        "reg",
        "epochs",
        "verbose"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.batch_size= self.conf["batch_size"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items
        self.p_dims = self.conf["p_dim"] + [self.num_items]
        self.q_dims = self.p_dims[::-1]
        self.dims = self.q_dims + self.p_dims[1:]
        self.act = self.conf["activation"]
        self.reg = self.conf["reg"]
        self.num_epochs=self.conf["epochs"]
        self.loss_function="multinominal-likelihood"
        self.verbose=self.conf["verbose"]

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.input_ph = tf.placeholder(dtype=tf.float32, shape=[None, self.num_items])
            self.keep_prob_ph = tf.placeholder_with_default(1.0, shape=None)

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            self.weights = []
            self.biases = []

            # define weights
            for i, (d_in, d_out) in enumerate(zip(self.dims[:-1], self.dims[1:])):
                weight_key = "weight_{}to{}".format(i, i+1)
                bias_key = "bias_{}".format(i+1)

                self.weights.append(tf.get_variable(
                    name=weight_key, shape=[d_in, d_out],
                    initializer=tf.contrib.layers.xavier_initializer()))

                self.biases.append(tf.get_variable(
                    name=bias_key, shape=[d_out],
                    initializer=tf.truncated_normal_initializer(
                        stddev=0.001)))

    def _create_inference(self):
        with tf.name_scope("inference"):
            # construct forward graph
            self.h = tf.nn.l2_normalize(self.input_ph, 1)
            self.h = tf.nn.dropout(self.h, self.keep_prob_ph)

            for i, (w, b) in enumerate(zip(self.weights, self.biases)):
                self.h = tf.matmul(self.h, w) + b

                if i != len(self.weights) - 1:
                    self.h = tool.activation_function(self.act, self.h)

            self.log_softmax_var = tf.nn.log_softmax(self.h)

    def _create_loss(self):
        with tf.name_scope("loss"):
            # per-user average negative log-likelihood
            neg_ll = -tf.reduce_mean(tf.reduce_sum(
            self.log_softmax_var * self.input_ph, axis=1))
            # apply regularization to weights
            regularization = l2_regularizer(self.reg)
            reg_var = apply_regularization(regularization, self.weights)
            # tensorflow l2 regularization multiply 0.5 to the l2 norm
            # multiply 2 so that it is back in the same scale
            self.loss = neg_ll + 2 * reg_var

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()

    def train_model(self):

        for epoch in  range(self.num_epochs):
            random_perm_doc_idx = np.random.permutation(self.num_users)
            self.total_batch = self.num_users
            total_loss = 0.0
            training_start_time = time()
            num_training_instances = self.num_users
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                if num_batch == self.total_batch - 1:
                    batch_set_idx = random_perm_doc_idx[num_batch * self.batch_size:]
                elif num_batch < self.total_batch - 1:
                    batch_set_idx = random_perm_doc_idx[num_batch * self.batch_size: (num_batch + 1) * self.batch_size]

                batch_matrix = np.zeros((len(batch_set_idx),self.num_items))

                batch_uid = 0
                trainDict = self.dataset.trainDict
                for userid in batch_set_idx:
                    items_by_userid = trainDict[userid]
                    for itemid in items_by_userid:
                        batch_matrix[batch_uid,itemid] = 1

                    batch_uid=batch_uid+1

                feed_dict = feed_dict={self.input_ph: batch_matrix,self.keep_prob_ph: 0.5}
                _, loss = self.sess.run([self.optimizer, self.loss],feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        rating_matrix = np.zeros((1,self.num_items), dtype=np.int32)
        items_by_userid = self.dataset.trainDict[user_id]
        for itemid in items_by_userid:
            rating_matrix[0,itemid] = 1
        output = self.sess.run(self.h, feed_dict={self.input_ph:rating_matrix})
        return output[0,items]



'''
Created on Apr 15, 2016
Evaluate the performance of Top-K recommendation:
    Protocol: leave-1-out evaluation
    Measures: Hit Ratio and NDCG
    (more details are in: Xiangnan He, et al. Fast Matrix Factorization for Online Recommendation with Implicit Feedback. SIGIR'16)
@author: wubin
'''
import numpy as np
from time import time
from neurec.evaluation.leaveoneout.LeaveOneOutEvaluate import evaluate_by_loo
from neurec.evaluation.foldout.FoldOutEvaluate import evaluate_by_foldout
import logging
def test_model(model,dataset,num_thread=10):
    eval_begin = time()
    model_name=str(model.__class__).split(sep=".")[-1].replace("\'>","")
    if dataset.splitter == "loo":
        (hits, ndcgs,aucs) = evaluate_by_loo(model,dataset.testMatrix,dataset.testNegatives,num_thread)
        hr = np.array(hits).mean()
        ndcg = np.array(ndcgs).mean()
        auc = np.array(aucs).mean()
        logging.info(
            "[model=%s]: [Test HR = %.6f, NDCG = %.6f,AUC = %.6f] [Time=%.1fs]" % (model_name,
            hr, ndcg,auc, time() - eval_begin))

    else:
        (pres,recs,maps,ndcgs,mrrs) = evaluate_by_foldout(model,dataset.testMatrix,dataset.testNegatives,num_thread)
        Precision = np.array(pres).mean()
        Recall = np.array(recs).mean()
        MAP = np.array(maps).mean()
        NDCG = np.array(ndcgs).mean()
        MRR = np.array(mrrs).mean()
        logging.info("[model=%s][%.1fs]: [Test Precision = %.6f, Recall= %.6f, MAP= %.6f, NDCG= %.6f, MRR= %.6f][topk=%.4s]"
               %(model_name,time() - eval_begin, Precision, Recall,MAP,NDCG,MRR,model.topK))



'''
Reference: Wu, Yao, et al. "Collaborative denoising auto-encoders for top-n recommender systems." in WSDM2016
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
from neurec.util.properties import Properties
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner,tool
from neurec.evaluation import Evaluate

class CDAE(AbstractRecommender):
    properties = [
        "hidden_neuron",
        "learning_rate",
        "learner",
        "loss_function",
        "topk",
        "reg",
        "epochs",
        "batch_size",
        "verbose",
        "h_act",
        "g_act",
        "corruption_level"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.hidden_neuron = self.conf["hidden_neuron"]
        self.learning_rate = self.conf["learning_rate"]
        self.learner = self.conf["learner"]
        self.loss_function = self.conf["loss_function"]
        self.topK = self.conf["topk"]
        self.reg = self.conf["reg"]
        self.num_epochs= self.conf["epochs"]
        self.batch_size= self.conf["batch_size"]
        self.verbose= self.conf["verbose"]
        self.h_act = self.conf["h_act"]
        self.g_act = self.conf["g_act"]
        self.corruption_level = self.conf["corruption_level"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape=[None,],name = 'user_input')
            self.input_R = tf.placeholder(tf.float32, [None, self.num_items])
            self.mask_corruption = tf.placeholder(tf.float32, [None, self.num_items])

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            self.V = tf.Variable(tf.random_normal([self.num_users,  self.hidden_neuron], stddev=0.01))

            self.weights = {
            'encoder': tf.Variable(tf.random_normal([self.num_items, self.hidden_neuron],stddev=0.01)),
            'decoder': tf.Variable(tf.random_normal([self.hidden_neuron, self.num_items],stddev=0.01)),
            }
            self.biases = {
            'encoder': tf.Variable(tf.random_normal([self.hidden_neuron],stddev=0.01)),
            'decoder': tf.Variable(tf.random_normal([self.num_items],stddev=0.01)),
            }

    def _create_inference(self):
        with tf.name_scope("inference"):

            self.user_latent =  tf.nn.embedding_lookup(self.V, self.user_input)

            corrupted_input = tf.multiply(self.input_R,self.mask_corruption)
            encoder_op = tool.activation_function(self.h_act,\
            tf.matmul(corrupted_input, self.weights['encoder'])+self.biases['encoder']+self.user_latent)

            self.decoder_op = tf.matmul(encoder_op, self.weights['decoder'])+self.biases['decoder']
            self.output = tool.activation_function(self.g_act,self.decoder_op)

    def _create_loss(self):
        with tf.name_scope("loss"):

            self.loss = - tf.reduce_sum(self.input_R* tf.log(self.output) + (1 - self.input_R) * tf.log(1 - self.output))

            self.reg_loss = self.reg*(tf.nn.l2_loss(self.weights['encoder'])+tf.nn.l2_loss(self.weights['decoder'])+
                tf.nn.l2_loss(self.biases['encoder'])+tf.nn.l2_loss(self.biases['decoder']))

            self.reg_loss = self.reg_loss + self.reg*tf.nn.l2_loss(self.user_latent)
            self.loss = self.loss + self.reg_loss
    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()

    def train_model(self):

        for epoch in  range(self.num_epochs):
            # Generate training instances
            mask_corruption_np = np.random.binomial(1, 1-self.corruption_level,
                                                (self.num_users, self.num_items))
            random_perm_doc_idx = np.random.permutation(self.num_users)
            self.total_batch = self.num_users
            total_loss = 0.0
            training_start_time = time()
            num_training_instances = self.num_users
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                if num_batch == self.total_batch - 1:
                    batch_set_idx = random_perm_doc_idx[num_batch * self.batch_size:]
                elif num_batch < self.total_batch - 1:
                    batch_set_idx = random_perm_doc_idx[num_batch * self.batch_size: (num_batch + 1) * self.batch_size]

                batch_matrix = np.zeros((len(batch_set_idx),self.num_items))

                batch_uid = 0
                trainDict = self.dataset.trainDict
                for userid in batch_set_idx:
                    items_by_userid = trainDict[userid]
                    for itemid in items_by_userid:
                        batch_matrix[batch_uid,itemid] = 1

                    batch_uid=batch_uid+1

                feed_dict = feed_dict={self.mask_corruption:
                    mask_corruption_np[batch_set_idx, :],\
                    self.input_R: batch_matrix, self.user_input: batch_set_idx}
                _, loss = self.sess.run([self.optimizer, self.loss],feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        mask = np.ones((1,self.num_items), dtype=np.int32)
        rating_matrix = np.zeros((1,self.num_items), dtype=np.int32)
        items_by_userid = self.dataset.trainDict[user_id]
        for itemid in items_by_userid:
            rating_matrix[0,itemid] = 1
        output = self.sess.run(self.output, feed_dict={self.mask_corruption:mask,self.input_R:rating_matrix,
            self.user_input: np.array([user_id])})
        return output[0,items]



# def hit(gt_items, pred_items):
#     count = 0
#     for item in pred_items:
#         if item in gt_items:
#             count += 1
#     return count

def getHitRatio(ranklist, gtItem):
    for item in ranklist:
        if item == gtItem:
            return 1
    return 0


import scipy.sparse as sp
import numpy as np
from neurec.util import reader
import logging

class GivenData(object):
    def __init__(self, path, dataset_name, separator,threshold):
        self.path =path
        self.dataset_name = dataset_name
        self.separator = separator
        self.threshold = threshold
        global num_items,num_users,userids,itemids,idusers,iditems

    def load_pre_splitter_data(self):
        pos_per_user={}
        num_items,num_users = 0,0
        userids,itemids,idusers,iditems = {},{},{},{}
        # Get number of users and items

        data = reader.lines(self.path + '/' + self.dataset_name)

        for line in data:
            useridx, itemidx, rating, time= line.strip().split(self.separator)
            if float(rating)>= self.threshold:
                if  itemidx not in itemids:
                    iditems[num_items]=itemidx
                    itemids[itemidx] = num_items
                    num_items+=1

                if useridx not in userids:
                    idusers[num_users]=useridx
                    userids[useridx]=num_users
                    num_users+=1
                    pos_per_user[userids[useridx]]=[]
                pos_per_user[userids[useridx]].append([itemids[itemidx],1,int(time)])
            else:
                if  itemidx not in itemids:
                    iditems[num_items]=itemidx
                    itemids[itemidx] = num_items
                    num_items+=1

                if useridx not in userids:
                    idusers[num_users]=useridx
                    userids[useridx]=num_users
                    num_users+=1
                    pos_per_user[userids[useridx]]=[]
                pos_per_user[userids[useridx]].append((itemids[itemidx],rating,int(time)))

            train_dict = {}
            for u in range(num_users):
                pos_per_user[u] =sorted(pos_per_user[u], key=lambda d: d[2])
                items = []
                for enlement in pos_per_user[u]:
                    items.append(enlement[0])
                train_dict[u] = items

        with open(self.path+".test.rating", 'r') as f:
            for line in f.readlines():
                useridx, itemidx,rating, time= line.strip().split(self.separator)
                if float(rating)>= self.threshold:
                    if  itemidx not in itemids:
                        iditems[num_items]=itemidx
                        itemids[itemidx] = num_items
                        num_items+=1

                    if useridx not in userids:
                        idusers[num_users]=useridx
                        userids[useridx]=num_users
                        num_users+=1
                        pos_per_user[userids[useridx]]=[]
                    pos_per_user[userids[useridx]].append([itemids[itemidx],1,int(time)])
                else :
                    if  itemidx not in itemids:
                        iditems[num_items]=itemidx
                        itemids[itemidx] = num_items
                        num_items+=1

                    if useridx not in userids:
                        idusers[num_users]=useridx
                        userids[useridx]=num_users
                        num_users+=1
                        pos_per_user[userids[useridx]]=[]
                    pos_per_user[userids[useridx]].append([itemids[itemidx],rating,int(time)])
        for u in range(num_users):
            pos_per_user[u]=sorted(pos_per_user[u], key=lambda d: d[2])

        train_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
        time_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
        with open(self.path+".train.rating", "r") as f:
            line = f.readline()
            while line != None and line != "":
                arr = line.split("\t")
                user, item, rating,time = userids[arr[0]], itemids[arr[1]], float(arr[2]), float(arr[3])
                if float(rating)>= self.threshold:
                    train_matrix[user, item] = 1

                else :
                    train_matrix[user, item] = rating
                time_matrix[user, item] = time
                line = f.readline()
        logging.info("already load the trainMatrix...")

        test_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
        with open(self.path+".test.rating", "r") as f:
            line = f.readline()
            while line != None and line != "":
                arr = line.split("\t")
                user, item, rating, time = userids[arr[0]], itemids[arr[1]], float(arr[2]), float(arr[3])
                if float(rating)>= self.threshold:
                    test_matrix[user, item] = 1
                else :
                    test_matrix[user, item] = rating
                time_matrix[user, item] = time
                line = f.readline()
        logging.info("already load the trainMatrix...")

        return train_matrix,train_dict,test_matrix,pos_per_user,userids,itemids,time_matrix



'''
Reference: Wang Xiang et al. "Neural Graph Collaborative Filtering." in SIGIR2019
Source code: https://github.com/xiangwang1223/neural_graph_collaborative_filtering
@author: wubin
'''
import tensorflow as tf
import numpy as np
import scipy.sparse as sp
from time import time
from neurec.util import data_gen
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
from neurec.util.properties import Properties

class NGCF(AbstractRecommender):
    properties = [
        "learning_rate",
        "learner",
        "topk",
        "batch_size",
        "embedding_size",
        "layer_size",
        "epochs",
        "reg",
        "loss_function",
        "node_dropout_flag",
        "adj_type",
        "alg_type",
        "verbose"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.batch_size= self.conf["batch_size"]
        self.emb_dim = self.conf["embedding_size"]
        self.weight_size = self.conf["layer_size"]
        self.n_layers = len(self.weight_size)
        self.num_epochs=self.conf["epochs"]
        self.decay=self.conf["reg"]
        self.loss_function=self.conf["loss_function"]
        self.node_dropout_flag = self.conf["node_dropout_flag"]
        self.adj_type = self.conf["adj_type"]
        self.alg_type = self.conf["alg_type"]
        self.n_fold = 100
        self.verbose=self.conf["verbose"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items
        self.graph = self.dataset.trainMatrix.toarray()
        self.norm_adj= self.get_adj_mat()
        self.n_nonzero_elems = self.norm_adj.count_nonzero()
        self.pretrain_data = None

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            # placeholder definition
            self.users = tf.placeholder(tf.int32, shape=(None,))
            self.pos_items = tf.placeholder(tf.int32, shape=(None,))
            self.neg_items = tf.placeholder(tf.int32, shape=(None,))

            # dropout: node dropout (adopted on the ego-networks);
            #          ... since the usage of node dropout have higher computational cost,
            #          ... please use the 'node_dropout_flag' to indicate whether use such technique.
            #          message dropout (adopted on the convolution operations).
            self.node_dropout = tf.placeholder_with_default([0.], shape=[None])
            self.mess_dropout = tf.placeholder_with_default([0.], shape=[None])

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            """
            *********************************************************
            Compute Graph-based Representations of all users & items via Message-Passing Mechanism of Graph Neural Networks.
            Different Convolutional Layers:
                1. ngcf: defined in 'Neural Graph Collaborative Filtering', SIGIR2019;
                2. gcn:  defined in 'Semi-Supervised Classification with Graph Convolutional Networks', ICLR2018;
                3. gcmc: defined in 'Graph Convolutional Matrix Completion', KDD2018;
            """
            # initialization of model parameters
            self.weights = self._init_weights()

            if self.alg_type in ['ngcf']:
                self.ua_embeddings, self.ia_embeddings = self._create_ngcf_embed()

            elif self.alg_type in ['gcn']:
                self.ua_embeddings, self.ia_embeddings = self._create_gcn_embed()

            elif self.alg_type in ['gcmc']:
                self.ua_embeddings, self.ia_embeddings = self._create_gcmc_embed()


    def _create_inference(self):
        with tf.name_scope("inference"):
            """
            *********************************************************
            Establish the final representations for user-item pairs in batch.
            """
            self.u_g_embeddings = tf.nn.embedding_lookup(self.ua_embeddings, self.users)
            self.pos_i_g_embeddings = tf.nn.embedding_lookup(self.ia_embeddings, self.pos_items)
            self.neg_i_g_embeddings = tf.nn.embedding_lookup(self.ia_embeddings, self.neg_items)

            """
            *********************************************************
            Inference for the testing phase.
            """
            self.batch_ratings = tf.matmul(self.u_g_embeddings, self.pos_i_g_embeddings, transpose_a=False, transpose_b=True)


    def _create_loss(self):
        with tf.name_scope("loss"):

            self.pos_scores = tf.reduce_sum(tf.multiply(self.u_g_embeddings, self.pos_i_g_embeddings), axis=1)
            neg_scores = tf.reduce_sum(tf.multiply(self.u_g_embeddings, self.neg_i_g_embeddings), axis=1)

            regularizer = tf.nn.l2_loss(self.u_g_embeddings) + tf.nn.l2_loss(self.pos_i_g_embeddings) + tf.nn.l2_loss(self.neg_i_g_embeddings)
            regularizer = regularizer/self.batch_size

            maxi = tf.log(tf.nn.sigmoid(self.pos_scores - neg_scores))
            mf_loss = tf.negative(tf.reduce_mean(maxi))

            emb_loss = self.decay * regularizer

            reg_loss = tf.constant(0.0, tf.float32, [1])

            self.loss = mf_loss + emb_loss + reg_loss
    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer=tf.train.AdamOptimizer(learning_rate=self.learning_rate).minimize(self.loss)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()

    def train_model(self):
        for epoch in  range(self.num_epochs):
            # Generate training instances
            user_input, item_input_pos, item_input_neg = data_gen._get_pairwise_all_data(self.dataset)

            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                bat_users,bat_items_pos,bat_items_neg =\
                 data_gen._get_pairwise_batch_data(user_input,\
                 item_input_pos, item_input_neg, num_batch, self.batch_size)
                feed_dict = {self.users: bat_users, self.pos_items: bat_items_pos,
                              self.node_dropout: [0.1],
                              self.mess_dropout: [0.1],
                              self.neg_items: bat_items_neg}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss

            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        users = np.full(len(items), user_id, dtype=np.int32)
        return self.sess.run(self.pos_scores, feed_dict={self.users: users, self.pos_items: items})

    def _create_ngcf_embed(self):
        # Generate a set of adjacency sub-matrix.
        if self.node_dropout_flag =='True':
            # node dropout.
            A_fold_hat = self._split_A_hat_node_dropout(self.norm_adj)
        else:
            A_fold_hat = self._split_A_hat(self.norm_adj)

        ego_embeddings = tf.concat([self.weights['user_embedding'], self.weights['item_embedding']], axis=0)

        all_embeddings = [ego_embeddings]

        for k in range(0, self.n_layers):

            temp_embed = []
            for f in range(self.n_fold):
                temp_embed.append(tf.sparse_tensor_dense_matmul(A_fold_hat[f], ego_embeddings))

            # sum messages of neighbors.
            side_embeddings = tf.concat(temp_embed, 0)
            # transformed sum messages of neighbors.
            sum_embeddings = tf.nn.leaky_relu(
                tf.matmul(side_embeddings, self.weights['W_gc_%d' % k]) + self.weights['b_gc_%d' % k])

            # bi messages of neighbors.
            bi_embeddings = tf.multiply(ego_embeddings, side_embeddings)
            # transformed bi messages of neighbors.
            bi_embeddings = tf.nn.leaky_relu(
                tf.matmul(bi_embeddings, self.weights['W_bi_%d' % k]) + self.weights['b_bi_%d' % k])

            # non-linear activation.
            ego_embeddings = sum_embeddings + bi_embeddings

            # message dropout.
            ego_embeddings = tf.nn.dropout(ego_embeddings, 1 - self.mess_dropout[k])

            # normalize the distribution of embeddings.
            norm_embeddings = tf.math.l2_normalize(ego_embeddings, axis=1)

            all_embeddings += [norm_embeddings]

        all_embeddings = tf.concat(all_embeddings, 1)
        u_g_embeddings, i_g_embeddings = tf.split(all_embeddings, [self.num_users, self.num_items], 0)
        return u_g_embeddings, i_g_embeddings

    def _create_gcn_embed(self):
        A_fold_hat = self._split_A_hat(self.norm_adj)
        embeddings = tf.concat([self.weights['user_embedding'], self.weights['item_embedding']], axis=0)


        all_embeddings = [embeddings]

        for k in range(0, self.n_layers):
            temp_embed = []
            for f in range(self.n_fold):
                temp_embed.append(tf.sparse_tensor_dense_matmul(A_fold_hat[f], embeddings))

            embeddings = tf.concat(temp_embed, 0)
            embeddings = tf.nn.leaky_relu(tf.matmul(embeddings, self.weights['W_gc_%d' %k]) + self.weights['b_gc_%d' %k])
            embeddings = tf.nn.dropout(embeddings, 1 - self.mess_dropout[k])

            all_embeddings += [embeddings]

        all_embeddings = tf.concat(all_embeddings, 1)
        u_g_embeddings, i_g_embeddings = tf.split(all_embeddings, [self.num_users, self.num_items], 0)
        return u_g_embeddings, i_g_embeddings

    def _create_gcmc_embed(self):
        A_fold_hat = self._split_A_hat(self.norm_adj)

        embeddings = tf.concat([self.weights['user_embedding'], self.weights['item_embedding']], axis=0)

        all_embeddings = []

        for k in range(0, self.n_layers):
            temp_embed = []
            for f in range(self.n_fold):
                temp_embed.append(tf.sparse_tensor_dense_matmul(A_fold_hat[f], embeddings))
            embeddings = tf.concat(temp_embed, 0)
            # convolutional layer.
            embeddings = tf.nn.leaky_relu(tf.matmul(embeddings, self.weights['W_gc_%d' % k]) + self.weights['b_gc_%d' % k])
            # dense layer.
            mlp_embeddings = tf.matmul(embeddings, self.weights['W_mlp_%d' %k]) + self.weights['b_mlp_%d' %k]
            mlp_embeddings = tf.nn.dropout(mlp_embeddings, 1 - self.mess_dropout[k])

            all_embeddings += [mlp_embeddings]
        all_embeddings = tf.concat(all_embeddings, 1)

        u_g_embeddings, i_g_embeddings = tf.split(all_embeddings, [self.num_users, self.num_items], 0)
        return u_g_embeddings, i_g_embeddings

    def _init_weights(self):
        all_weights = dict()

        initializer = tf.contrib.layers.xavier_initializer()

        if self.pretrain_data is None:
            all_weights['user_embedding'] = tf.Variable(initializer([self.num_users, self.emb_dim]), name='user_embedding')
            all_weights['item_embedding'] = tf.Variable(initializer([self.num_items, self.emb_dim]), name='item_embedding')
            self.logger.info('using xavier initialization')
        else:
            all_weights['user_embedding'] = tf.Variable(initial_value=self.pretrain_data['user_embed'], trainable=True,
                                                        name='user_embedding', dtype=tf.float32)
            all_weights['item_embedding'] = tf.Variable(initial_value=self.pretrain_data['item_embed'], trainable=True,
                                                        name='item_embedding', dtype=tf.float32)
            self.logger.info('using pretrained initialization')

        self.weight_size_list = [self.emb_dim] + self.weight_size

        for k in range(self.n_layers):
            all_weights['W_gc_%d' %k] = tf.Variable(
                initializer([self.weight_size_list[k], self.weight_size_list[k+1]]), name='W_gc_%d' % k)
            all_weights['b_gc_%d' %k] = tf.Variable(
                initializer([1, self.weight_size_list[k+1]]), name='b_gc_%d' % k)

            all_weights['W_bi_%d' % k] = tf.Variable(
                initializer([self.weight_size_list[k], self.weight_size_list[k + 1]]), name='W_bi_%d' % k)
            all_weights['b_bi_%d' % k] = tf.Variable(
                initializer([1, self.weight_size_list[k + 1]]), name='b_bi_%d' % k)

            all_weights['W_mlp_%d' % k] = tf.Variable(
                initializer([self.weight_size_list[k], self.weight_size_list[k+1]]), name='W_mlp_%d' % k)
            all_weights['b_mlp_%d' % k] = tf.Variable(
                initializer([1, self.weight_size_list[k+1]]), name='b_mlp_%d' % k)

        return all_weights

    def normalized_adj_single(self,adj):
        rowsum = np.array(adj.sum(1))

        d_inv = np.power(rowsum, -1).flatten()
        d_inv[np.isinf(d_inv)] = 0.
        d_mat_inv = sp.diags(d_inv)
        norm_adj = d_mat_inv.dot(adj)
        # norm_adj = adj.dot(d_mat_inv)
        self.logger.info('generate single-normalized adjacency matrix.')
        return norm_adj.tocoo()
    def get_adj_mat(self):
        A = sp.dok_matrix((self.num_users + self.num_items, self.num_users + self.num_items), dtype=np.float32)
        A = A.tolil()
        A[:self.num_users, self.num_users:] = self.graph
        A[self.num_users:, :self.num_users] = self.graph.transpose()
        A = A.todok()
        if self.adj_type== 'plain':
            adj_mat = A
            self.logger.info('use the plain adjacency matrix')
        elif self.adj_type == 'norm':
            adj_mat = self.normalized_adj_single(A + sp.eye(A.shape[0]))
            self.logger.info('use the normalized adjacency matrix')
        elif self.adj_type == 'gcmc':
            adj_mat = self.normalized_adj_single(A)
            self.logger.info('use the gcmc adjacency matrix')
        else:
            adj_mat = self.normalized_adj_single(A) + sp.eye(A.shape[0])
            self.logger.info('use the mean adjacency matrix')

        return adj_mat.tocsr()

    def _split_A_hat(self, X):
        A_fold_hat = []

        fold_len = (self.num_users + self.num_items) // self.n_fold
        for i_fold in range(self.n_fold):
            start = i_fold * fold_len
            if i_fold == self.n_fold -1:
                end = self.num_users + self.num_items
            else:
                end = (i_fold + 1) * fold_len

            A_fold_hat.append(self._convert_sp_mat_to_sp_tensor(X[start:end]))
        return A_fold_hat

    def _split_A_hat_node_dropout(self, X):
        A_fold_hat = []

        fold_len = (self.num_users + self.num_items) // self.n_fold
        for i_fold in range(self.n_fold):
            start = i_fold * fold_len
            if i_fold == self.n_fold -1:
                end = self.num_users + self.num_items
            else:
                end = (i_fold + 1) * fold_len

            # A_fold_hat.append(self._convert_sp_mat_to_sp_tensor(X[start:end]))
            temp = self._convert_sp_mat_to_sp_tensor(X[start:end])
            n_nonzero_temp = X[start:end].count_nonzero()
            A_fold_hat.append(self._dropout_sparse(temp, 1 - self.node_dropout[0], n_nonzero_temp))
        return A_fold_hat

    def _dropout_sparse(self, X, keep_prob, n_nonzero_elems):
        """
        Dropout for sparse tensors.
        """
        noise_shape = [n_nonzero_elems]
        random_tensor = keep_prob
        random_tensor += tf.random_uniform(noise_shape)
        dropout_mask = tf.cast(tf.floor(random_tensor), dtype=tf.bool)
        pre_out = tf.sparse_retain(X, dropout_mask)

        return pre_out * tf.div(1., keep_prob)

    def _convert_sp_mat_to_sp_tensor(self, X):
        coo = X.tocoo().astype(np.float32)
        indices = np.mat([coo.row, coo.col]).transpose()
        return tf.SparseTensor(indices, coo.data, coo.shape)



'''
Reference: Tong Zhao et al., "Leveraging Social Connections to Improve
Personalized Ranking for Collaborative Filtering." in CIKM 2014
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import scipy.sparse as sp
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
import logging
from neurec.util.properties import Properties
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
def random_choice(a, size=None, replace=True, p=None, exclusion=None):
    # TODO exclusion is element, not the index
    if exclusion is not None:
        if p is None:
            p = np.ones_like(a)
        else:
            p = np.array(p, copy=True)
        p = np.ndarray.flatten(p)
        p[exclusion] = 0
        p = p / np.sum(p)
    sample = np.random.choice(a, size=size, replace=replace, p=p)
    return sample

class SBPR(AbstractRecommender):
    properties = [
        "socialpath",
        "learning_rate",
        "embedding_size",
        "learner",
        "loss_function",
        "topk",
        "num_epochs",
        "batch_size",
        "verbose"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.socialpath = self.conf["socialpath"]
        self.learning_rate = self.conf["learning_rate"]
        self.embedding_size = self.conf["embedding_size"]
        self.learner = self.conf["learner"]
        self.loss_function = self.conf["loss_function"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["num_epochs"]
        self.reg_mf = self.conf["reg_mf"]
        self.batch_size = self.conf["batch_size"]
        self.verbose = self.conf["verbose"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items
        self.userids = self.dataset.userids
        self.dataset_name = self.dataset.dataset_name
        self.userouterids = self.userids.keys()
        trainMatrix = self.dataset.trainMatrix.tocsr()
        self.train_dict = {u: set(pos_item.indices) for u, pos_item in enumerate(trainMatrix)}
        self.socialMatrix=self._get_social_data()
        self.userSocialItemsSetList = self._get_SocialItemsSet_sun()

    def _get_social_data(self):
        social_users = np.genfromtxt(self.socialpath, dtype=None, names=["user0", "user1"], delimiter=',')
        users_key = np.array(list(self.userids.keys()))
        user0 = social_users["user0"].astype(np.str)
        index = np.in1d(user0, users_key)
        social_users = social_users[index]

        user1 = social_users["user1"].astype(np.str)
        index = np.in1d(user1, users_key)
        social_users = social_users[index]

        user0 = social_users["user0"].astype(np.str)
        user0_id = [self.userids[u] for u in user0]
        user1 = social_users["user1"].astype(np.str)
        user1_id = [self.userids[u] for u in user1]
        social_matrix = sp.csr_matrix(([1]*len(user0_id), (user0_id, user1_id)),
                                      shape=(self.num_users, self.num_users))
        return social_matrix

    def _get_SocialItemsSet_sun(self):
        #find items rated by trusted neighbors only
        userSocialItemsSetList = {}
        for u, _ in self.train_dict.items():
            trustors = self.socialMatrix[u].indices
            items = [item for f_u in trustors for item in self.train_dict[f_u] if item not in self.train_dict[u]]
            items = set(items)
            if len(items) > 0:
                userSocialItemsSetList[u] = list(items)
        return userSocialItemsSetList

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape = [None,], name = "user_input")
            self.item_input_pos = tf.placeholder(tf.int32, shape = [None,], name = "item_input_pos")
            self.item_input_social = tf.placeholder(tf.int32, shape = [None,], name = "item_input_social")
            self.suk = tf.placeholder(tf.float32, shape = [None,], name = "suk")
            self.item_input_neg = tf.placeholder(tf.int32, shape = [None,], name = "item_input_neg")
    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.user_embeddings = tf.Variable(tf.random_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01),
                                                                name='user_embeddings', dtype=tf.float32)  #(users, embedding_size)
            self.item_embeddings = tf.Variable(tf.random_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                                                                name='item_embeddings', dtype=tf.float32)  #(items, embedding_size)
            self.bias = tf.Variable(tf.zeros(self.num_items),name='bias')
    def _create_inference(self, item_input):
        with tf.name_scope("inference"):
            # embedding look up
            user_embedding = tf.nn.embedding_lookup(self.user_embeddings, self.user_input)
            item_embedding = tf.nn.embedding_lookup(self.item_embeddings, item_input)
            item_bias = tf.nn.embedding_lookup(self.bias, item_input)
            output = tf.reduce_sum(tf.multiply(user_embedding, item_embedding),1) + item_bias
            return user_embedding, item_embedding,item_bias, output

    def _create_loss(self):
        with tf.name_scope("loss"):
            # loss for L(Theta)
            p1, q1,b1,self.output= self._create_inference(self.item_input_pos)
            _, q2,b2, output_social = self._create_inference(self.item_input_social)
            _, q3,b3, output_neg = self._create_inference(self.item_input_neg)
            result1 = tf.divide(self.output - output_social,self.suk)
            result2 = output_social - output_neg
            self.loss = learner.pairwise_loss(self.loss_function,result1)+learner.pairwise_loss(self.loss_function,result2)+ self.reg_mf * ( tf.reduce_sum(tf.square(p1)) \
            + tf.reduce_sum(tf.square(q2)) + tf.reduce_sum(tf.square(q1))+tf.reduce_sum(tf.square(q3))+tf.reduce_sum(tf.square(b1))+tf.reduce_sum(tf.square(b2))+tf.reduce_sum(tf.square(b3)))
    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()
#---------- training process -------
    def train_model(self):
        for epoch in range(self.num_epochs):
            # Generate training instances
#             logging.info("get training data")
            user_input, item_input_pos,item_input_social,item_input_neg,suk_input = self._get_pairwise_all_data_sun()
#             logging.info("begin training")
            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
#                 self.logger.info(num_batch)
                num_training_instances =len(user_input)
                id_start = num_batch * self.batch_size
                id_end = (num_batch + 1) *self.batch_size
                if id_end>num_training_instances:
                    id_end=num_training_instances
                bat_users = user_input[id_start:id_end]
                bat_items_pos = item_input_pos[id_start:id_end]
                bat_items_social = item_input_social[id_start:id_end]
                bat_items_neg = item_input_neg[id_start:id_end]
                bat_suk_input = suk_input[id_start:id_end]
                feed_dict = {self.user_input:bat_users,self.item_input_pos:bat_items_pos,\
                            self.item_input_social:bat_items_social,\
                            self.item_input_neg:bat_items_neg,self.suk:bat_suk_input}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def _get_pairwise_all_data(self):
        user_input, item_input_pos,item_input_social,item_input_neg,suk_input = [],[],[],[],[]
        trainMatrix = self.dataset.trainMatrix
        num_items = self.dataset.num_items
        for (u, i) in trainMatrix.keys():
            if u in self.userSocialItemsSetList:
                user_input.append(u)
                item_input_pos.append(i)
                socialItemsList = self.userSocialItemsSetList[u]
                j = np.random.randint(num_items)
                while (u, j) in trainMatrix.keys() or j in socialItemsList:
                    j = np.random.randint(num_items)
                item_input_neg.append(j)
                k = np.random.choice(socialItemsList)
                item_input_social.append(k)
                trustorIndices = self.socialMatrix[u].indices
                socialWeight = 0
                for trustedUserIdx in trustorIndices:
                    indices = trainMatrix[trustedUserIdx].tocsr().indices
                    if k in indices:
                        socialWeight += 1
                suk_input.append(socialWeight+1)
        user_input = np.array(user_input, dtype=np.int32)
        item_input_pos = np.array(item_input_pos, dtype=np.int32)
        item_input_social = np.array(item_input_social, dtype=np.int32)
        item_input_neg = np.array(item_input_neg, dtype=np.int32)
        suk_input = np.array(suk_input, dtype=np.float32)
        num_training_instances = len(user_input)
        shuffle_index = np.arange(num_training_instances,dtype=np.int32)
        np.random.shuffle(shuffle_index)
        user_input=user_input[shuffle_index]
        item_input_pos=item_input_pos[shuffle_index]
        item_input_social=item_input_social[shuffle_index]
        item_input_neg=item_input_neg[shuffle_index]
        suk_input = suk_input[shuffle_index]
        return user_input, item_input_pos,item_input_social,item_input_neg,suk_input

    def _get_pairwise_all_data_sun(self):
        user_input, item_input_pos, item_input_social, item_input_neg, suk_input = [],[],[],[],[]

        num_items = self.dataset.num_items
        all_items = np.arange(num_items)

        for u, pos_item in self.train_dict.items():
            if u not in self.userSocialItemsSetList:
                continue
            # pos_item = pos_item.indices.tolist()
            pos_len = len(pos_item)
            user_input.extend([u]*pos_len)
            item_input_pos.extend(pos_item)

            socialItemsList = self.userSocialItemsSetList[u]
            # a, size = None, replace = True, p = None, exclusion = None
            neg_excl = np.concatenate([socialItemsList, list(pos_item)], axis=0)

            neg_item = random_choice(all_items, pos_len, replace=True, exclusion=neg_excl)
            item_input_neg.extend(neg_item)
            social_item = np.random.choice(socialItemsList, size=pos_len)
            item_input_social.extend(social_item)

            trustedUserIdices = self.socialMatrix[u].indices
            socialWeight_bool = [[1 if k in self.train_dict[f_u] else 0 for f_u in trustedUserIdices]
                                 for k in social_item]
            socialWeight = np.sum(socialWeight_bool, axis=-1) + 1
            suk_input.extend(socialWeight)

        user_input = np.array(user_input, dtype=np.int32)
        item_input_pos = np.array(item_input_pos, dtype=np.int32)
        item_input_social = np.array(item_input_social, dtype=np.int32)
        item_input_neg = np.array(item_input_neg, dtype=np.int32)
        suk_input = np.array(suk_input, dtype=np.float32)
        num_training_instances = len(user_input)
        shuffle_index = np.arange(num_training_instances,dtype=np.int32)
        np.random.shuffle(shuffle_index)
        user_input=user_input[shuffle_index]
        item_input_pos=item_input_pos[shuffle_index]
        item_input_social=item_input_social[shuffle_index]
        item_input_neg=item_input_neg[shuffle_index]
        suk_input = suk_input[shuffle_index]
        return user_input, item_input_pos,item_input_social,item_input_neg,suk_input

    def predict(self, user_id, eval_items):
        users = np.full(len(eval_items), user_id, dtype=np.int32)
        return self.sess.run(self.output, feed_dict={self.user_input: users, self.item_input_pos: eval_items})



'''
Reference: Hong-Jian Xue et al., "Deep Matrix Factorization Models for Recommender Systems." In IJCAI2017.
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties

class DMF(AbstractRecommender):
    properties = [
        "embedding_size",
        "learning_rate",
        "learner",
        "topk",
        "epochs",
        "num_neg",
        "batch_size",
        "verbose",
        "layers",
        "loss_function"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.embedding_size = self.conf["embedding_size"]
        self.learning_rate = self.conf["learning_rate"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.num_negatives= self.conf["num_neg"]
        self.batch_size= self.conf["batch_size"]
        self.verbose= self.conf["verbose"]
        self.layers = self.conf["layers"]
        self.loss_function = self.conf["loss_function"]
        self.fist_layer_size = self.layers[0]
        self.last_layer_size = self.layers[1]
        self.neg_sample_size = self.num_negatives
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items
        self.user_matrix = self.dataset.trainMatrix.tocsr()
        self.item_matrix = self.dataset.trainMatrix.tocsc()
        self.dataset_name = self.dataset.dataset_name

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.one_hot_u = tf.placeholder(tf.float32, shape=[None,None],name = 'user_input')
            self.one_hot_v = tf.placeholder(tf.float32, shape=[None,None],name = 'item_input')
            self.lables = tf.placeholder(tf.float32, shape=[None,],name="labels")

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            self.u_w1 = tf.get_variable("u_w1", shape=(self.num_items, self.fist_layer_size), initializer=tf.contrib.layers.xavier_initializer())
            self.u_b1 = tf.get_variable("u_b1", shape=[self.fist_layer_size], initializer=tf.contrib.layers.xavier_initializer())
            self.u_w2 = tf.get_variable("u_w2", shape=(self.fist_layer_size, self.last_layer_size), initializer=tf.contrib.layers.xavier_initializer())
            self.u_b2 = tf.get_variable("u_b2", shape=[self.last_layer_size], initializer=tf.contrib.layers.xavier_initializer())

            self.v_w1 = tf.get_variable("v_w1", shape=(self.num_users, self.fist_layer_size), initializer=tf.contrib.layers.xavier_initializer())
            self.v_b1 = tf.get_variable("v_b1", shape=[self.fist_layer_size], initializer=tf.contrib.layers.xavier_initializer())
            self.v_w2 = tf.get_variable("v_w2", shape=(self.fist_layer_size, self.last_layer_size), initializer=tf.contrib.layers.xavier_initializer())
            self.v_b2 = tf.get_variable("v_b2", shape=[self.last_layer_size], initializer=tf.contrib.layers.xavier_initializer())

    def _create_inference(self):
        with tf.name_scope("inference"):
            net_u_1 = tf.nn.relu(tf.matmul(self.one_hot_u, self.u_w1) + self.u_b1)
            net_u_2 = tf.matmul(net_u_1, self.u_w2) + self.u_b2

            net_v_1 = tf.nn.relu(tf.matmul(self.one_hot_v, self.v_w1) + self.v_b1)
            net_v_2 = tf.matmul(net_v_1, self.v_w2) + self.v_b2

            fen_zhi = tf.reduce_sum(net_u_2 * net_v_2, 1)

            norm_u = tf.reduce_sum(tf.square(net_u_2), 1)
            norm_v = tf.reduce_sum(tf.square(net_v_2), 1)
            fen_mu = norm_u * norm_v
            self.output =  tf.nn.relu(fen_zhi / fen_mu)

    def _create_loss(self):
        with tf.name_scope("loss"):
            self.loss = learner.pointwise_loss(self.loss_function, self.lables, self.output)
    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()

    def train_model(self):

        for epoch in  range(self.num_epochs):
            # Generate training instances
            user_input, item_input, lables = self._get_input_all_data()

            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                num_training_instances =len(user_input)
                id_start = num_batch * self.batch_size
                id_end = (num_batch + 1) *  self.batch_size
                if id_end>num_training_instances:
                    id_end=num_training_instances
                bat_users = user_input[id_start:id_end].tolist()
                bat_items = item_input[id_start:id_end].tolist()
                bat_lables = np.array(lables[id_start:id_end])
                feed_dict = {self.one_hot_u:bat_users, self.one_hot_v:bat_items,
                             self.lables:bat_lables}
                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)
    def _get_input_all_data(self):
        user_input,item_input,lables = [],[],[]
        for u in range(self.num_users):
            # positive instance
            items_by_user = self.user_matrix[u].indices
            u_vector = np.reshape(self.user_matrix.getrow(u).toarray(),[self.num_items])
            for i in items_by_user:
                i_vector = np.reshape(self.item_matrix.getcol(i).toarray(),[self.num_users])
                user_input.append(u_vector)
                item_input.append(i_vector)
                lables.append(1)
            # negative instance
                for _ in range(self.num_negatives):
                    j = np.random.randint(self.num_items)
                    while (u, j) in self.dataset.trainMatrix.keys():
                        j = np.random.randint(self.num_items)
                    j_vector = np.reshape(self.item_matrix.getcol(i).toarray(),[self.num_users])
                    user_input.append(u_vector)
                    item_input.append(j_vector)
                    lables.append(0)
        user_input = np.array(user_input, dtype=np.int32)
        item_input = np.array(item_input, dtype=np.int32)
        lables = np.array(lables, dtype=np.float32)
        num_training_instances = len(user_input)
        shuffle_index = np.arange(num_training_instances,dtype=np.int32)
        np.random.shuffle(shuffle_index)
        user_input=user_input[shuffle_index]
        item_input=item_input[shuffle_index]
        lables = lables[shuffle_index]
        return user_input,item_input,lables
    def predict(self, user_id, items):
        user_input,item_input = [],[]
        u_vector = np.reshape(self.user_matrix.getrow(user_id).toarray(),[self.num_items])
        for i in items:
            user_input.append(u_vector)
            i_vector = np.reshape(self.item_matrix.getcol(i).toarray(),[self.num_users])
            item_input.append(i_vector)
        return self.sess.run(self.output, feed_dict={self.one_hot_u: user_input, self.one_hot_v: item_input})



import tensorflow as tf
def optimizer(learner,loss,learning_rate,momentum=0.9):
    optimizer=None
    if learner.lower() == "adagrad":
        optimizer = tf.train.AdagradOptimizer(learning_rate=learning_rate,\
                     initial_accumulator_value=1e-8).minimize(loss)
    elif learner.lower() == "rmsprop":
        optimizer = tf.train.RMSPropOptimizer(learning_rate).minimize(loss)
    elif learner.lower() == "adam":
        optimizer = tf.train.AdamOptimizer(learning_rate).minimize(loss)
    elif learner.lower() == "gd" :
        optimizer = tf.train.GradientDescentOptimizer(learning_rate).minimize(loss)
    elif learner.lower() == "momentum" :
        optimizer = tf.train.MomentumOptimizer(learning_rate,momentum).minimize(loss)
    else :
        raise ValueError("please select a suitable optimizer")
    return optimizer

def pairwise_loss(loss_function,y,margin=1):
    loss=None
    if loss_function.lower() == "bpr":
        loss = -tf.reduce_sum(tf.log_sigmoid(y))
    elif loss_function.lower() == "hinge":
        loss = tf.reduce_sum(tf.maximum(y+margin, 0))
    elif loss_function.lower() == "square":
        loss = tf.reduce_sum(tf.square(1-y))
    else:
        raise Exception("please choose a suitable loss function")
    return loss

def pointwise_loss(loss_function,y_rea,y_pre):
    loss=None
    if loss_function.lower() == "cross_entropy":
        loss = tf.losses.sigmoid_cross_entropy(y_rea,y_pre)
#         loss = - tf.reduce_sum(
#             y_rea * tf.log(y_pre) + (1 - y_rea) * tf.log(1 - y_pre))
    elif loss_function.lower() == "square":
        loss = tf.reduce_sum(tf.square(y_rea-y_pre))
    else:
        raise Exception("please choose a suitable loss function")
    return loss



'''
Reference: Pengfei Wang et al., "Learning Hierarchical Representation Model for NextBasket Recommendation." in SIGIR 2015.
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner, data_gen
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
class HRM(AbstractRecommender):
    properties = [
        "learning_rate",
        "embedding_size",
        "learner",
        "topk",
        "epochs",
        "reg_mf",
        "pre_agg",
        "loss_function",
        "session_agg",
        "batch_size",
        "high_order",
        "verbose",
        "num_neg"
    ]
    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.embedding_size = self.conf["embedding_size"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.reg_mf = self.conf["reg_mf"]
        self.pre_agg= self.conf["pre_agg"]
        self.loss_function = self.conf["loss_function"]
        self.session_agg= self.conf["session_agg"]
        self.batch_size= self.conf["batch_size"]
        self.high_order = self.conf["high_order"]
        self.verbose= self.conf["verbose"]
        self.num_negatives= self.conf["num_neg"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape = [None,], name = "user_input")
            self.item_input = tf.placeholder(tf.int32, shape = [None,], name = "item_input_pos")
            self.item_input_recent = tf.placeholder(tf.int32, shape = [None,None], name = "item_input_recent")
            self.lables = tf.placeholder(tf.float32, shape=[None,],name="labels")
    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.user_embeddings = tf.Variable(tf.truncated_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01),
                name='user_embeddings', dtype=tf.float32)  #(users, embedding_size)
            self.item_embeddings = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='item_embeddings', dtype=tf.float32)  #(items, embedding_size)

    def avg_pooling(self, imput_embedding):
        output = tf.reduce_mean(imput_embedding, axis=1)
        return output

    def max_pooling(self, imput_embedding):
        output = tf.reduce_max(imput_embedding, axis=1)
        return output

    def _create_inference(self):
        with tf.name_scope("inference"):
            # embedding look up
            user_embedding = tf.nn.embedding_lookup(self.user_embeddings, self.user_input)
            item_embedding_recent = tf.nn.embedding_lookup(self.item_embeddings, self.item_input_recent)
            item_embedding = tf.nn.embedding_lookup(self.item_embeddings, self.item_input)
            if self.high_order>1:
                if self.session_agg == "max":
                    item_embedding_short = self.max_pooling(item_embedding_recent)
                elif self.session_agg == "avg":
                    item_embedding_short = self.avg_pooling(item_embedding_recent)
                concat_user_item = tf.concat([tf.expand_dims(user_embedding,1),tf.expand_dims(item_embedding_short,1)],axis=1)
            else:
                concat_user_item = tf.concat([tf.expand_dims(user_embedding,1),tf.expand_dims(item_embedding_recent,1)],axis=1)
            if self.pre_agg == "max":
                hybrid_user_embedding = self.max_pooling(concat_user_item)
            elif self.pre_agg == "avg":
                hybrid_user_embedding = self.avg_pooling(concat_user_item)
            predict_vector = tf.multiply(hybrid_user_embedding, item_embedding)
            predict = tf.reduce_sum(predict_vector,1)
            return user_embedding, item_embedding,item_embedding_recent,predict


    def _create_loss(self):
        with tf.name_scope("loss"):
            # loss for L(Theta)
            p1,q1,r1,self.output = self._create_inference()
            self.loss = learner.pointwise_loss(self.loss_function,self.lables,self.output) + self.reg_mf * (tf.reduce_sum(tf.square(p1)) \
            +tf.reduce_sum(tf.square(r1))+ tf.reduce_sum(tf.square(q1)))

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()
#---------- training process -------
    def train_model(self):

        for epoch in  range(self.num_epochs):
            # Generate training instances
            user_input, item_input,item_input_recents, lables = data_gen._get_pointwise_all_highorder_data(self.dataset,self.high_order, self.num_negatives)

            num_training_instances = len(user_input)
            total_loss = 0.0
            training_start_time = time()

            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                bat_users, bat_items, bat_items_recents, bat_lables =\
                data_gen._get_pointwise_batch_seqdata(user_input, \
                item_input,item_input_recents, lables, num_batch, self.batch_size)
                feed_dict = {self.user_input:bat_users, self.item_input:bat_items,
                             self.item_input_recent:bat_items_recents,self.lables:bat_lables}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss

            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)
    def predict(self, user_id, items):
        cand_items = self.dataset.trainDict[user_id]
        item_recents = []
        for _ in range(len(items)):
            item_recents.append( cand_items[len(cand_items)-self.high_order:])
        users = np.full(len(items), user_id, dtype='int32')
        return self.sess.run((self.output), feed_dict={self.user_input: users,\
                                        self.item_input_recent:item_recents, self.item_input: items})



"""Functions to handle reading files in the package."""
from configparser import ConfigParser, MissingSectionHeaderError
import logging

def file(path):
    """Returns the configuration settings for a file.

    path -- path to the file
    """
    parser = ConfigParser()

    try:
        with open(path) as file:
            parser.read_file(file)
    except FileNotFoundError:
        raise FileNotFoundError("Could not find file " + str(path) + ". Make sure the file path is correct.")
    except MissingSectionHeaderError:
        raise RuntimeError("Could not find a section header in " + str(path) + '. Added [DEFAULT] to line 1 of this file')

    return parser

def lines(file_path):
    """Returns all lines from a file.

    file_path -- path of the file to read
    """
    try:
        with open(file_path) as file:
            return file.readlines()
    except FileNotFoundError:
        raise FileNotFoundError(str(file_path) + " could not be found. Check the file path is correct.")



'''
Reference: Xiangnan He et al., "NAIS: Neural Attentive Item Similarity Model for Recommendation." in TKDE2018
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division

import os
from neurec.model.AbstractRecommender import AbstractRecommender

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.evaluation import Evaluate
from neurec.util import data_gen,learner
from neurec.util.properties import Properties

class NAIS(AbstractRecommender):
    properties = [
        "pretrain",
        "verbose",
        "batch_size",
        "epochs",
        "weight_size",
        "embedding_size",
        "data_alpha",
        "regs",
        "ispairwise",
        "topk",
        "alpha",
        "beta",
        "num_neg",
        "learning_rate",
        "activation",
        "loss_function",
        "learner",
        "algorithm"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.pretrain = self.conf["pretrain"]
        self.verbose = self.conf["verbose"]
        self.batch_size = self.conf["batch_size"]
        self.num_epochs = self.conf["epochs"]
        self.weight_size = self.conf["weight_size"]
        self.embedding_size = self.conf["embedding_size"]
        self.data_alpha = self.conf["data_alpha"]
        self.regs = self.conf["regs"]
        self.ispairwise = self.conf["ispairwise"]
        self.topK = self.conf["topk"]
        self.lambda_bilinear = self.regs[0]
        self.gamma_bilinear = self.regs[1]
        self.eta_bilinear = self.regs[2]
        self.alpha = self.conf["alpha"]
        self.beta = self.conf["beta"]
        self.num_negatives= self.conf["num_neg"]
        self.learning_rate = self.conf["learning_rate"]
        self.activation = self.conf["activation"]
        self.loss_function = self.conf["loss_function"]
        self.algorithm = self.conf["algorithm"]
        self.learner = self.conf["learner"]
        self.dataset_name = self.dataset.dataset_name
        self.num_items = self.dataset.num_items
        self.num_users = self.dataset.num_users

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape=[None, None], name = "user_input")    #the index of users
            self.num_idx = tf.placeholder(tf.float32, shape=[None,],name = "num_idx")    #the number of items rated by users
            self.item_input = tf.placeholder(tf.int32, shape=[None,],name = "item_input_pos")      #the index of items
            if self.ispairwise == True:
                self.user_input_neg = tf.placeholder(tf.int32, shape=[None, None], name = "user_input_neg")
                self.item_input_neg = tf.placeholder(tf.int32, shape = [None,], name = "item_input_neg")
                self.num_idx_neg = tf.placeholder(tf.float32, shape=[None,],name = "num_idx_neg")
            else :
                self.lables = tf.placeholder(tf.float32, shape=[None,],name="labels")

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            trainable_flag = (self.pretrain!=2)
            self.c1 = tf.Variable(tf.truncated_normal(shape=[self.num_items,\
                self.embedding_size], mean=0.0, stddev=0.01), name='c1', \
                    dtype=tf.float32, trainable=trainable_flag)

            self.c2 = tf.constant(0.0, tf.float32, [1, self.embedding_size], name='c2')
            self.embedding_Q_ = tf.concat([self.c1, self.c2], axis=0, name='embedding_Q_')
            self.embedding_Q = tf.Variable(tf.truncated_normal(shape=[self.num_items,\
                self.embedding_size], mean=0.0, stddev=0.01),name='embedding_Q',\
                dtype=tf.float32,trainable=trainable_flag)

            self.bias = tf.Variable(tf.zeros(self.num_items),name='bias',trainable=trainable_flag)

            # Variables for attention
            if self.algorithm == 0:
                self.W = tf.Variable(tf.truncated_normal(shape=[self.embedding_size,\
                 self.weight_size], mean=0.0, stddev=tf.sqrt(tf.div(2.0, self.weight_size\
                 + self.embedding_size))),name='Weights_for_MLP', dtype=tf.float32, trainable=True)
            else:
                self.W = tf.Variable(tf.truncated_normal(shape=[2*self.embedding_size,\
                self.weight_size], mean=0.0, stddev=tf.sqrt(tf.div(2.0, self.weight_size\
                + 2*self.embedding_size))),name='Weights_for_MLP', dtype=tf.float32, trainable=True)

            self.b = tf.Variable(tf.truncated_normal(shape=[1, self.weight_size],\
                mean=0.0, stddev=tf.sqrt(tf.div(2.0, self.weight_size + self.embedding_size))),\
                name='Bias_for_MLP', dtype=tf.float32, trainable=True)

            self.h = tf.Variable(tf.ones([self.weight_size, 1]), name='H_for_MLP', dtype=tf.float32)

    def _create_inference(self,user_input,item_input,num_idx):
        with tf.name_scope("inference"):
            embedding_q_ = tf.nn.embedding_lookup(self.embedding_Q_, user_input) # (b, n, e)
            embedding_q = tf.expand_dims(tf.nn.embedding_lookup(self.embedding_Q,item_input),1) # (b, 1, e)

            if self.algorithm == 0:
                embedding_p = self._attention_MLP(embedding_q_ * embedding_q,embedding_q_,num_idx)
            else:
                n = tf.shape(user_input)[1]
                embedding_p = self._attention_MLP(tf.concat([embedding_q_,\
                     tf.tile(embedding_q, tf.stack([1,n,1]))],2),embedding_q_,num_idx)

            embedding_q = tf.reduce_sum(embedding_q, 1)
            bias_i = tf.nn.embedding_lookup(self.bias, item_input)
            coeff = tf.pow(num_idx, tf.constant(self.alpha, tf.float32, [1]))
            output = coeff *tf.reduce_sum(embedding_p*embedding_q, 1)+ bias_i

            return embedding_q_,embedding_q,output

    def _create_loss(self):
        with tf.name_scope("loss"):

            p1, q1, self.output = self._create_inference(self.user_input,self.item_input,self.num_idx)
            if self.ispairwise == True:
                _, q2,output_neg = self._create_inference(self.user_input_neg,self.item_input_neg,self.num_idx_neg)
                self.result = self.output - output_neg
                self.loss = learner.pairwise_loss(self.loss_function,self.result) + self.lambda_bilinear * ( tf.reduce_sum(tf.square(p1))) \
                +self.gamma_bilinear*(tf.reduce_sum(tf.square(q2)) + tf.reduce_sum(tf.square(q1)))

            else:
                self.loss = learner.pointwise_loss(self.loss_function, \
                self.lables,self.output) + self.lambda_bilinear *\
                (tf.reduce_sum(tf.square(p1)))+self.gamma_bilinear *(tf.reduce_sum(tf.square(q1)))

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()

    def _attention_MLP(self, q_,embedding_q_,num_idx):
            with tf.name_scope("attention_MLP"):
                b = tf.shape(q_)[0]
                n = tf.shape(q_)[1]
                r = (self.algorithm + 1)*self.embedding_size

                MLP_output = tf.matmul(tf.reshape(q_,[-1,r]), self.W) + self.b #(b*n, e or 2*e) * (e or 2*e, w) + (1, w)
                if self.activation == 0:
                    MLP_output = tf.nn.relu( MLP_output )
                elif self.activation == 1:
                    MLP_output = tf.nn.sigmoid( MLP_output )
                elif self.activation == 2:
                    MLP_output = tf.nn.tanh( MLP_output )

                A_ = tf.reshape(tf.matmul(MLP_output, self.h),[b,n]) #(b*n, w) * (w, 1) => (None, 1) => (b, n)

                # softmax for not mask features
                exp_A_ = tf.exp(A_)
                mask_mat = tf.sequence_mask(num_idx, maxlen = n, dtype = tf.float32) # (b, n)
                exp_A_ = mask_mat * exp_A_
                exp_sum = tf.reduce_sum(exp_A_, 1, keepdims=True)  # (b, 1)
                exp_sum = tf.pow(exp_sum, tf.constant(self.beta, tf.float32, [1]))

                A = tf.expand_dims(tf.div(exp_A_, exp_sum),2) # (b, n, 1)

                return tf.reduce_sum(A * embedding_q_, 1)

    def train_model(self):

        for epoch in  range(self.num_epochs):
            if self.ispairwise == True:
                user_input,user_input_neg, num_idx_pos, num_idx_neg, item_input_pos,item_input_neg = \
                data_gen._get_pairwise_all_likefism_data(self.dataset)
            else :
                user_input,num_idx,item_input,lables = data_gen._get_pointwise_all_likefism_data(self.dataset,self.num_negatives)

            num_training_instances = len(user_input)
            total_loss = 0.0
            training_start_time = time()
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                if self.ispairwise == True:
                    bat_users_pos,bat_users_neg, bat_idx_pos, bat_idx_neg, \
                        bat_items_pos,bat_items_neg= \
                        data_gen._get_pairwise_batch_likefism_data(user_input,\
                        user_input_neg,self.dataset.num_items, num_idx_pos, num_idx_neg, item_input_pos,\
                        item_input_neg, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users_pos,self.user_input_neg:bat_users_neg,\
                                self.num_idx:bat_idx_pos,self.num_idx_neg:bat_idx_neg,
                                self.item_input:bat_items_pos,self.item_input_neg:bat_items_neg}
                else :
                    bat_users,bat_idx,bat_items,bat_lables =\
                        data_gen._get_pointwise_batch_likefism_data(user_input,self.dataset.num_items,
                        num_idx,item_input,lables, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users,self.num_idx:bat_idx, self.item_input:bat_items,
                                self.lables:bat_lables}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id,items):
        cand_items = self.dataset.trainDict[user_id]
        num_idx = len(cand_items)
        # Get prediction scores
        item_idx = np.full(len(items), num_idx, dtype=np.int32)
        user_input = []
        for _ in range(len(items)):
            user_input.append(cand_items)
        feed_dict = {self.user_input: np.array(user_input), \
                     self.num_idx: item_idx, self.item_input:items}
        return self.sess.run((self.output), feed_dict=feed_dict)



'''
Reference: Ziwei Zhu, et al. "Improving Top-K Recommendation via Joint
Collaborative Autoencoders." in WWW2019
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner,tool
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties

class JCA(AbstractRecommender):
    properties = [
        "hidden_neuron",
        "learning_rate",
        "learner",
        "topk",
        "reg",
        "epochs",
        "batch_size",
        "verbose",
        "f_act",
        "g_act",
        "margin",
        "corruption_level",
        "num_neg"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.hidden_neuron = self.conf["hidden_neuron"]
        self.learning_rate = self.conf["learning_rate"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.reg = self.conf["reg"]
        self.num_epochs= self.conf["epochs"]
        self.batch_size= self.conf["batch_size"]
        self.verbose= self.conf["verbose"]
        self.f_act = self.conf["f_act"] # the activation function for the output layer
        self.g_act = self.conf["g_act"] # the activation function for the hidden layer
        self.margin = self.conf["margin"]
        self.corruption_level = self.conf["corruption_level"]
        self.neg_sample_rate = self.conf["num_neg"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

        self.train_R = self.dataset.trainMatrix.toarray()
        self.U_OH_mat = np.eye(self.num_users, dtype=float)
        self.I_OH_mat = np.eye(self.num_items, dtype=float)
        self.num_batch_U = int(self.num_users / float(self.batch_size)) + 1
        self.num_batch_I = int(self.num_items / float(self.batch_size)) + 1

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            # input rating vector
            self.input_R_U = tf.placeholder(dtype=tf.float32, shape=[None, self.num_items], name="input_R_U")
            self.input_R_I = tf.placeholder(dtype=tf.float32, shape=[self.num_users, None], name="input_R_I")
            self.input_OH_I = tf.placeholder(dtype=tf.float32, shape=[None, self.num_items], name="input_OH_I")
            self.input_P_cor = tf.placeholder(dtype=tf.int32, shape=[None, 2], name="input_P_cor")
            self.input_N_cor = tf.placeholder(dtype=tf.int32, shape=[None, 2], name="input_N_cor")

            # input indicator vector indicator
            self.row_idx = tf.placeholder(dtype=tf.int32, shape=[None, 1], name="row_idx")
            self.col_idx = tf.placeholder(dtype=tf.int32, shape=[None, 1], name="col_idx")

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            # user component
            # first layer weights
            self.UV = tf.get_variable(name="UV", initializer=tf.truncated_normal(shape=[self.num_items, self.hidden_neuron],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)
            # second layer weights
            self.UW = tf.get_variable(name="UW", initializer=tf.truncated_normal(shape=[self.hidden_neuron, self.num_items],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)
            # first layer bias
            self.Ub1 = tf.get_variable(name="Ub1", initializer=tf.truncated_normal(shape=[1, self.hidden_neuron],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)
            # second layer bias
            self.Ub2 = tf.get_variable(name="Ub2", initializer=tf.truncated_normal(shape=[1, self.num_items],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)

            # item component
            # first layer weights
            self.IV = tf.get_variable(name="IV", initializer=tf.truncated_normal(shape=[self.num_users, self.hidden_neuron],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)
            # second layer weights
            self.IW = tf.get_variable(name="IW", initializer=tf.truncated_normal(shape=[self.hidden_neuron, self.num_users],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)
            # first layer bias
            self.Ib1 = tf.get_variable(name="Ib1", initializer=tf.truncated_normal(shape=[1, self.hidden_neuron],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)
            # second layer bias
            self.Ib2 = tf.get_variable(name="Ib2", initializer=tf.truncated_normal(shape=[1, self.num_users],
                                                                            mean=0, stddev=0.03), dtype=tf.float32)


            self.I_factor_vector = tf.get_variable(name="I_factor_vector", initializer=tf.random_uniform(shape=[1, self.num_items]),
                                              dtype=tf.float32)

    def _create_inference(self):
        with tf.name_scope("inference"):

            # user component
            U_pre_Encoder = tf.matmul(self.input_R_U, self.UV) + self.Ub1  # input to the hidden layer
            self.U_Encoder = tool.activation_function(self.g_act,U_pre_Encoder)  # output of the hidden layer
            U_pre_Decoder = tf.matmul(self.U_Encoder, self.UW) + self.Ub2  # input to the output layer
            self.U_Decoder = tool.activation_function(self.f_act,U_pre_Decoder)  # output of the output layer

            # item component
            I_pre_mul = tf.transpose(tf.matmul(self.I_factor_vector, tf.transpose(self.input_OH_I)))
            I_pre_Encoder = tf.matmul(tf.transpose(self.input_R_I), self.IV) + self.Ib1  # input to the hidden layer
            self.I_Encoder = tool.activation_function(self.g_act,I_pre_Encoder * I_pre_mul)  # output of the hidden layer
            I_pre_Decoder = tf.matmul(self.I_Encoder, self.IW) + self.Ib2  # input to the output layer
            self.I_Decoder = tool.activation_function(self.f_act,I_pre_Decoder)  # output of the output layer

            # final output
            self.Decoder = ((tf.transpose(tf.gather_nd(tf.transpose(self.U_Decoder), self.col_idx)))
                            + tf.gather_nd(tf.transpose(self.I_Decoder), self.row_idx)) / 2.0

            pos_data = tf.gather_nd(self.Decoder, self.input_P_cor)
            neg_data = tf.gather_nd(self.Decoder, self.input_N_cor)

            self.pre_cost1 = tf.maximum(neg_data - pos_data + self.margin,
                                   tf.zeros(tf.shape(neg_data)[0]))

    def _create_loss(self):
        with tf.name_scope("loss"):

            cost1 = tf.reduce_sum(self.pre_cost1)  # prediction squared error
            pre_cost2 = tf.square(self.l2_norm(self.UW)) + tf.square(self.l2_norm(self.UV)) \
                        + tf.square(self.l2_norm(self.IW)) + tf.square(self.l2_norm(self.IV))\
                        + tf.square(self.l2_norm(self.Ib1)) + tf.square(self.l2_norm(self.Ib2))\
                        + tf.square(self.l2_norm(self.Ub1)) + tf.square(self.l2_norm(self.Ub2))
            cost2 = self.reg * 0.5 * pre_cost2  # regularization term

            self.cost = cost1 + cost2  # the loss function
    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.cost, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()

    def train_model(self):

        for epoch in  range(self.num_epochs):
            random_row_idx = np.random.permutation(self.num_users)  # randomly permute the rows
            random_col_idx = np.random.permutation(self.num_items)  # randomly permute the cols
            training_start_time = time()
            total_loss = 0.0
            for i in range(self.num_batch_U):  # iterate each batch
                if i == self.num_batch_U - 1:
                    row_idx = random_row_idx[i * self.batch_size:]
                else:
                    row_idx = random_row_idx[(i * self.batch_size):((i + 1) * self.batch_size)]
                for j in range(self.num_batch_I):
                    # get the indices of the current batch
                    if j == self.num_batch_I - 1:
                        col_idx = random_col_idx[j * self.batch_size:]
                    else:
                        col_idx = random_col_idx[(j * self.batch_size):((j + 1) * self.batch_size)]

                    p_input, n_input = self.pairwise_neg_sampling(row_idx, col_idx)
                    input_tmp = self.train_R[row_idx, :]
                    input_tmp = input_tmp[:, col_idx]

                    input_R_U = self.train_R[row_idx, :]
                    input_R_I = self.train_R[:, col_idx]
                    _, loss = self.sess.run(  # do the optimization by the minibatch
                        [self.optimizer, self.cost],
                        feed_dict={
                            self.input_R_U: input_R_U,
                            self.input_R_I: input_R_I,
                            self.input_OH_I: self.I_OH_mat[col_idx, :],
                            self.input_P_cor: p_input,
                            self.input_N_cor: n_input,
                            self.row_idx: np.reshape(row_idx, (len(row_idx), 1)),
                            self.col_idx: np.reshape(col_idx, (len(col_idx), 1))})
                total_loss+=loss
            self.logger.info("[iter %d : total_loss : %f, time: %f]" %(epoch+1,total_loss,time()-training_start_time))
            if epoch %self.verbose == 0:
                self.eval_rating_matrix()
                Evaluate.test_model(self,self.dataset)

    def pairwise_neg_sampling(self,row_idx, col_idx):
        R = self.train_R[row_idx, :]
        R = R[:, col_idx]
        p_input, n_input = [], []
        obsv_list = np.where(R == 1)

        unobsv_mat = []
        for r in range(R.shape[0]):
            unobsv_list = np.where(R[r, :] == 0)
            unobsv_list = unobsv_list[0]
            unobsv_mat.append(unobsv_list)

        for i in range(len(obsv_list[1])):
            # positive instance
            u = obsv_list[0][i]
            # negative instances
            unobsv_list = unobsv_mat[u]
            neg_samp_list = np.random.choice(unobsv_list, size=self.neg_sample_rate, replace=False)
            for ns in neg_samp_list:
                p_input.append([u, obsv_list[1][i]])
                n_input.append([u, ns])
        # self.logger.info('dataset size = ' + str(len(p_input)))
        return np.array(p_input), np.array(n_input)

    def eval_rating_matrix(self):
        self.allRatings = self.sess.run(self.Decoder,
            feed_dict={
            self.input_R_U: self.train_R,
            self.input_R_I: self.train_R,
            self.input_OH_I: self.I_OH_mat,
            self.input_P_cor: [[0, 0]],
            self.input_N_cor: [[0, 0]],
            self.row_idx: np.reshape(range(self.num_users), (self.num_users, 1)),
            self.col_idx: np.reshape(range(self.num_items), (self.num_items, 1))})

    def predict(self, user_id, items):
        return self.allRatings[user_id, items]

    def l2_norm(self,tensor):
        return tf.sqrt(tf.reduce_sum(tf.square(tensor)))



'''
Reference: Feng Xue et al., "Deep Item-based Collaborative Filtering for Top-N Recommendation" in TOIS2019
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf
import numpy as np
from time import time
from neurec.evaluation import Evaluate
from neurec.util import learner
from tensorflow.contrib.layers.python.layers import batch_norm as batch_norm
from neurec.util.properties import Properties

class DeepICF(AbstractRecommender):
    properties = [
        "pretrain",
        "verbose",
        "batch_choice",
        "batch_size",
        "batch_norm",
        "epochs",
        "weight_size",
        "embedding_size",
        "layers",
        "regs",
        "regw",
        "ispairwise",
        "topk",
        "alpha",
        "beta",
        "num_neg",
        "learning_rate",
        "activation",
        "loss_function",
        "algorithm",
        "learner"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.pretrain = self.conf["pretrain"]
        self.verbose = self.conf["verbose"]
        self.batch_choice = self.conf["batch_choice"]
        self.batch_size = self.conf["batch_size"]
        self.use_batch_norm = self.conf["batch_norm"]
        self.num_epochs = self.conf["epochs"]
        self.weight_size = self.conf["weight_size"]
        self.embedding_size = self.conf["embedding_size"]
        self.n_hidden = self.conf["layers"]
        self.regs = self.conf["regs"]
        self.reg_W = self.conf["regw"]
        self.ispairwise = self.conf["ispairwise"]
        self.topK = self.conf["topk"]
        self.lambda_bilinear = self.regs[0]
        self.gamma_bilinear = self.regs[1]
        self.eta_bilinear = self.regs[2]
        self.alpha = self.conf["alpha"]
        self.beta = self.conf["beta"]
        self.num_negatives= self.conf["num_neg"]
        self.learning_rate = self.conf["learning_rate"]
        self.activation = self.conf["activation"]
        self.loss_function = self.conf["loss_function"]
        self.algorithm = self.conf["algorithm"]
        self.learner = self.conf["learner"]
        self.num_items = self.dataset.num_items
        self.num_users = self.dataset.num_users

    # batch norm
    def batch_norm_layer(self,x, train_phase, scope_bn):
        bn_train = batch_norm(x, decay=0.9, center=True, scale=True, updates_collections=None,
            is_training=True, reuse=None, trainable=True, scope=scope_bn)
        bn_inference = batch_norm(x, decay=0.9, center=True, scale=True, updates_collections=None,
            is_training=False, reuse=True, trainable=True, scope=scope_bn)
        z = tf.cond(train_phase, lambda: bn_train, lambda: bn_inference)
        return z


    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape=[None, None])  # the index of users
            self.num_idx = tf.placeholder(tf.float32, shape=[None,])  # the number of items rated by users
            self.item_input = tf.placeholder(tf.int32, shape=[None,])  # the index of items
            self.labels = tf.placeholder(tf.float32, shape=[None,])  # the ground truth
            self.is_train_phase = tf.placeholder(tf.bool)  # mark is training or testing

    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            trainable_flag = (self.pretrain != 2)
            self.c1 = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01), name='c1', dtype=tf.float32, trainable=trainable_flag)
            self.c2 = tf.constant(0.0, tf.float32, [1, self.embedding_size], name='c2')
            self.embedding_Q_ = tf.concat([self.c1, self.c2], 0, name='embedding_Q_')
            self.embedding_Q = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01), name='embedding_Q', dtype=tf.float32, trainable=trainable_flag)
            self.bias = tf.Variable(tf.zeros(self.num_items), name='bias', trainable=trainable_flag)

            # Variables for attention
            if self.algorithm == 0:
                self.W = tf.Variable(tf.truncated_normal(shape=[self.embedding_size, self.weight_size], mean=0.0, stddev=tf.sqrt(tf.div(2.0, self.weight_size + self.embedding_size))), name='Weights_for_MLP', dtype=tf.float32, trainable=True)
            else:
                self.W = tf.Variable(tf.truncated_normal(shape=[2 * self.embedding_size, self.weight_size], mean=0.0, stddev=tf.sqrt(tf.div(2.0, self.weight_size + ( 2 * self.embedding_size)))), name='Weights_for_MLP', dtype=tf.float32, trainable=True)
            self.b = tf.Variable(tf.truncated_normal(shape=[1, self.weight_size], mean=0.0, stddev=tf.sqrt(tf.div(2.0, self.weight_size + self.embedding_size))), name='Bias_for_MLP', dtype=tf.float32, trainable=True)
            self.h = tf.Variable(tf.ones([self.weight_size, 1]), name='H_for_MLP', dtype=tf.float32)

            # Variables for DeepICF+a
            self.weights = {
                'out': tf.Variable(tf.random_normal([self.n_hidden[-1], 1], mean=0, stddev=np.sqrt(2.0 / (self.n_hidden[-1] + 1))), name='weights_out')
            }
            self.biases = {
                'out': tf.Variable(tf.random_normal([1]), name='biases_out')
            }
            n_hidden_0 = self.embedding_size
            for i in range(len(self.n_hidden)):
                if i > 0:
                    n_hidden_0 = self.n_hidden[i - 1]
                n_hidden_1 = self.n_hidden[i]
                self.weights['h%d' % i] = tf.Variable(tf.random_normal([n_hidden_0, n_hidden_1], mean=0, stddev=np.sqrt(2.0 / (n_hidden_0 + n_hidden_1))), name='weights_h%d' % i)
                self.biases['b%d' % i] = tf.Variable(tf.random_normal([n_hidden_1]), name='biases_b%d' % i)


    def _attention_MLP(self, q_):
        with tf.name_scope("attention_MLP"):
            b = tf.shape(q_)[0]
            n = tf.shape(q_)[1]
            r = (self.algorithm + 1) * self.embedding_size

            MLP_output = tf.matmul(tf.reshape(q_, [-1, r]), self.W) + self.b  # (b*n, e or 2*e) * (e or 2*e, w) + (1, w)
            if self.activation == 0:
                MLP_output = tf.nn.relu(MLP_output)
            elif self.activation == 1:
                MLP_output = tf.nn.sigmoid(MLP_output)
            elif self.activation == 2:
                MLP_output = tf.nn.tanh(MLP_output)

            A_ = tf.reshape(tf.matmul(MLP_output, self.h), [b, n])  # (b*n, w) * (w, 1) => (None, 1) => (b, n)

            # softmax for not mask features
            exp_A_ = tf.exp(A_)
            num_idx = self.num_idx
            mask_mat = tf.sequence_mask(num_idx, maxlen=n, dtype=tf.float32)  # (b, n)
            exp_A_ = mask_mat * exp_A_
            exp_sum = tf.reduce_sum(exp_A_, 1, keepdims=True)  # (b, 1)
            exp_sum = tf.pow(exp_sum, tf.constant(self.beta, tf.float32, [1]))

            A = tf.expand_dims(tf.div(exp_A_, exp_sum), 2)  # (b, n, 1)

            return A, tf.reduce_sum(A * self.embedding_q_, 1)

    def _create_inference(self):
        with tf.name_scope("inference"):
            self.embedding_q_ = tf.nn.embedding_lookup(self.embedding_Q_, self.user_input)  # (b, n, e)
            self.embedding_q = tf.nn.embedding_lookup(self.embedding_Q, self.item_input)  # (b, 1, e)

            if self.algorithm == 0:  # prod
                self.A, self.embedding_p = self._attention_MLP(self.embedding_q_ * tf.expand_dims(self.embedding_q,1))  # (?, k)
            else:  # concat
                n = tf.shape(self.user_input)[1]
                self.A, self.embedding_p = self._attention_MLP(tf.concat([self.embedding_q_, tf.tile(tf.expand_dims(self.embedding_q,1), tf.stack([1, n, 1]))], 2))  # (?, k)

            self.bias_i = tf.nn.embedding_lookup(self.bias, self.item_input)
            self.coeff = tf.pow(tf.expand_dims(self.num_idx,1), tf.constant(self.alpha, tf.float32, [1]))
            self.embedding_p = self.coeff * self.embedding_p  # (?, k)

            # DeepICF+a
            layer1 = tf.multiply(self.embedding_p, self.embedding_q)  # (?, k)
            for i in range(0,len(self.n_hidden)):
                layer1 = tf.add(tf.matmul(layer1, self.weights['h%d' % i]), self.biases['b%d' % i])
                if self.use_batch_norm:
                    layer1 = self.batch_norm_layer(layer1, train_phase=self.is_train_phase, scope_bn='bn_%d' % i)
                layer1 = tf.nn.relu(layer1)
            out_layer = tf.reduce_sum(tf.matmul(layer1, self.weights['out']) + self.biases['out'],1) # (?, 1)

            self.output = tf.sigmoid(tf.add_n([out_layer, self.bias_i]))  # (?, 1)

    def _create_loss(self):
        with tf.name_scope("loss"):
            self.loss = tf.losses.log_loss(self.labels, self.output) + \
                        self.lambda_bilinear * tf.reduce_sum(tf.square(self.embedding_Q)) + \
                        self.gamma_bilinear * tf.reduce_sum(tf.square(self.embedding_Q_)) + \
                        self.eta_bilinear * tf.reduce_sum(tf.square(self.W))

            for i in range(min(len(self.n_hidden), len(self.reg_W))):
                if self.reg_W[i] > 0:
                    self.loss = self.loss + self.reg_W[i] * tf.reduce_sum(tf.square(self.weights['h%d'%i]))
    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()
    def batch_gen(self,batches, i):
        return [(batches[r])[i] for r in range(4)]
    def train_model(self):

        for epoch in  range(self.num_epochs):
            batches = self.shuffle()
            num_batch = len(batches[1])
            batch_index = np.arange(num_batch)
            training_start_time = time()
            total_loss = 0.0
            for index in batch_index:
                user_input, num_idx, item_input, labels = self.batch_gen(batches, index)
                feed_dict = {self.user_input: user_input, self.num_idx: num_idx, self.item_input: item_input,
                             self.labels: labels, self.is_train_phase: True}
                loss,_ = self.sess.run([self.loss, self.optimizer], feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_batch,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def shuffle(self):   #negative sampling and shuffle the data
        if self.batch_choice == 'user':
            _user_input, _item_input, _labels, _batch_length = self._get_train_data_user()
            self._num_batch = len(_batch_length)
            user_input_list, num_idx_list, item_input_list, labels_list = [], [], [], []
            for i in range(int(self._num_batch)):
                ui, ni, ii, l = self._get_train_batch_user(i,_user_input, _item_input, _labels, _batch_length)
                user_input_list.append(ui)
                num_idx_list.append(ni)
                item_input_list.append(ii)
                labels_list.append(l)
            return user_input_list, num_idx_list, item_input_list, labels_list
        else:
            _user_input, _item_input, _labels = self._get_train_data_fixed()
            iterations = len(_user_input)
            self.index = np.arange(iterations)
            self._num_batch = iterations / self.batch_size
            user_input_list, num_idx_list, item_input_list, labels_list = [], [], [], []
            for i in range(int(self._num_batch)):
                ui, ni, ii, l = self._get_train_batch_fixed(i,_user_input, _item_input, _labels)
                user_input_list.append(ui)
                num_idx_list.append(ni)
                item_input_list.append(ii)
                labels_list.append(l)
            return user_input_list, num_idx_list, item_input_list, labels_list

    def _get_train_data_user(self):
        user_input, item_input, labels, batch_length = [],[],[],[]
        trainList = self.dataset.trainDict
        for u in range(self.num_users):
            if u == 0:
                batch_length.append((1+self.num_negatives) * len(trainList[u]))
            else:
                batch_length.append((1+self.num_negatives) * len(trainList[u])+batch_length[u-1])
            for i in trainList[u]:
                # positive instance
                user_input.append(u)
                item_input.append(i)
                labels.append(1)
                # negative instances
                for _ in range(self.num_negatives):
                    j = np.random.randint(self.num_items)
                    while j in trainList[u]:
                        j = np.random.randint(self.num_items)
                    user_input.append(u)
                    item_input.append(j)
                    labels.append(0)
        return  user_input, item_input, labels, batch_length


    def _get_train_batch_user(self,i,user_input, item_input, labels, batch_length):
        #represent the feature of users via items rated by him/her
        user_list, num_list, item_list, labels_list = [],[],[],[]
        trainList = self.dataset.trainDict
        if i == 0:
            begin = 0
        else:
            begin = batch_length[i-1]
        batch_index = list(range(begin, batch_length[i]))
        np.random.shuffle(batch_index)
        for idx in batch_index:
            user_idx = user_input[idx]
            item_idx = item_input[idx]
            nonzero_row = []
            nonzero_row += trainList[user_idx]
            num_list.append(self._remove_item(self.num_items, nonzero_row, item_idx))
            user_list.append(nonzero_row)
            item_list.append(item_idx)
            labels_list.append(labels[idx])
        user_input = np.array(self._add_mask(self.num_items, user_list, max(num_list)))
        num_idx = np.array(num_list)
        item_input = np.array(item_list)
        labels = np.array(labels_list)
        return (user_input, num_idx, item_input, labels)

    def _get_train_data_fixed(self):
        user_input, item_input, labels = [],[],[]
        train = self.dataset.trainMatrix
        for (u, i) in train.keys():
            # positive instance
            user_input.append(u)
            item_input.append(i)
            labels.append(1)
            # negative instances
            for _ in range(self.num_negatives):
                j = np.random.randint(self.num_items)
                while (u, j) in train.keys():
                # while train.has_key((u, j)):
                    j = np.random.randint(self.num_items)
                user_input.append(u)
                item_input.append(j)
                labels.append(0)
        return user_input, item_input, labels

    def _get_train_batch_fixed(self,i,user_input, item_input, labels):
        #represent the feature of users via items rated by him/her
        user_list, num_list, item_list,labels_list = [],[],[],[]
        trainList = self.dataset.trainList
        begin = i * self.batch_size
        for idx in range(begin, begin+self.batch_size):
            user_idx = user_input[self.index[idx]]
            item_idx = item_input[self.index[idx]]
            nonzero_row = []
            nonzero_row += trainList[user_idx]
            num_list.append(self._remove_item(self.num_items, nonzero_row, item_idx))
            user_list.append(nonzero_row)
            item_list.append(item_idx)
            labels_list.append(labels[self.index[idx]])
        user_input = np.array(self._add_mask(self.num_items, user_list, max(num_list)))
        num_idx = np.array(num_list)
        item_input = np.array(item_list)
        labels = np.array(labels_list)
        return (user_input, num_idx, item_input, labels)

    def _remove_item(self,feature_mask, users, item):
        flag = 0
        for i in range(len(users)):
            if users[i] == item:
                users[i] = users[-1]
                users[-1] = feature_mask
                flag = 1
                break
        return len(users) - flag

    def _add_mask(self,feature_mask, features, num_max):
        #uniformalize the length of each batch
        for i in range(len(features)):
            features[i] = features[i] + [feature_mask] * (num_max+1 - len(features[i]))
        return features

    def predict(self, user_id,items):
        cand_items = self.dataset.trainDict[user_id]
        num_idx = len(cand_items)
        # Get prediction scores
        item_idx = np.full(len(items), num_idx, dtype=np.int32)
        user_input = []
        for _ in range(len(items)):
            user_input.append(cand_items)
        feed_dict = {self.user_input: np.array(user_input), \
                     self.num_idx: item_idx, self.item_input:items,self.is_train_phase: False}
        return self.sess.run((self.output), feed_dict=feed_dict)



'''
Reference: Steffen Rendle et al., "BPR: Bayesian Personalized Ranking from Implicit Feedback." in UAI 2009.
    GMF: Xiangnan He et al., "Neural Collaborative Filtering." in WWW 2017.
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner,data_gen
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
from neurec.util.properties import Properties
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

class MF(AbstractRecommender):
    properties = [
        "learning_rate",
        "embedding_size",
        "learner",
        "loss_function",
        "ispairwise",
        "topk",
        "epochs",
        "reg_mf",
        "batch_size",
        "verbose",
        "num_neg"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.embedding_size = self.conf["embedding_size"]
        self.learner = self.conf["learner"]
        self.loss_function = self.conf["loss_function"]
        self.ispairwise = self.conf["ispairwise"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.reg_mf = self.conf["reg_mf"]
        self.batch_size= self.conf["batch_size"]
        self.verbose= self.conf["verbose"]
        self.num_negatives= self.conf["num_neg"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape = [None,], name = "user_input")
            self.item_input = tf.placeholder(tf.int32, shape = [None,], name = "item_input")
            if self.ispairwise == True:
                self.item_input_neg = tf.placeholder(tf.int32, shape = [None,], name = "item_input_neg")
            else :
                self.lables = tf.placeholder(tf.float32, shape=[None,],name="labels")
    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.user_embeddings = tf.Variable(tf.random_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01),
                                                                name='user_embeddings', dtype=tf.float32)  #(users, embedding_size)
            self.item_embeddings = tf.Variable(tf.random_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                                                                name='item_embeddings', dtype=tf.float32)  #(items, embedding_size)
    def _create_inference(self, item_input):
        with tf.name_scope("inference"):
            # embedding look up
            user_embedding = tf.nn.embedding_lookup(self.user_embeddings, self.user_input)
            item_embedding = tf.nn.embedding_lookup(self.item_embeddings, item_input)
            predict = tf.reduce_sum(tf.multiply(user_embedding, item_embedding),1)
            return user_embedding, item_embedding, predict

    def _create_loss(self):
        with tf.name_scope("loss"):
            # loss for L(Theta)
            p1, q1,self.output = self._create_inference(self.item_input)
            if self.ispairwise == True:
                _, q2, self.output_neg = self._create_inference(self.item_input_neg)
                result = self.output - self.output_neg
                self.loss = learner.pairwise_loss(self.loss_function,result) + self.reg_mf * ( tf.reduce_sum(tf.square(p1)) \
                + tf.reduce_sum(tf.square(q2)) + tf.reduce_sum(tf.square(q1)))

            else :
                self.loss = learner.pointwise_loss(self.loss_function, self.lables,self.output)+self.reg_mf * (tf.reduce_sum(tf.square(p1)) \
                   + tf.reduce_sum(tf.square(q1)))

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()
#---------- training process -------
    def train_model(self):

        for epoch in  range(self.num_epochs):
            # Generate training instances
            if self.ispairwise == True:
                user_input, item_input_pos, item_input_neg = data_gen._get_pairwise_all_data(self.dataset)
            else :
                user_input, item_input, lables = data_gen._get_pointwise_all_data(self.dataset, self.num_negatives)

            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                if self.ispairwise == True:
                    bat_users,bat_items_pos,bat_items_neg =\
                     data_gen._get_pairwise_batch_data(user_input,\
                     item_input_pos, item_input_neg, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users,self.item_input:bat_items_pos,\
                                self.item_input_neg:bat_items_neg}
                else:
                    bat_users, bat_items,bat_lables =\
                     data_gen._get_pointwise_batch_data(user_input, \
                     item_input, lables, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users, self.item_input:bat_items,
                                 self.lables:bat_lables}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)
                #Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        users = np.full(len(items), user_id, dtype=np.int32)
        return self.sess.run(self.output, feed_dict={self.user_input: users, self.item_input: items})
